"""Include functions for saving various results to hdf5."""

import logging
import os
from numbers import Integral
from typing import Union

import h5py
import numpy as np

from ..result import Result, SampleResult
from .hdf5 import write_array, write_float_array

logger = logging.getLogger(__name__)


def check_overwrite(
    f: Union[h5py.File, h5py.Group], overwrite: bool, target: str
):
    """
    Check whether target already exists.

    Sends a warning if ``overwrite=False``, deletes the target if
    ``overwrite=True``.

    Attributes
    ----------
    f: file or group where existence of a group with the path group_path
       should be checked
    target: name of the group, whose existence is checked
    overwrite: if ``True``, it deletes the target in ``f``
    """
    if target in f:
        if overwrite:
            del f[target]
        else:
            raise RuntimeError(
                f"File `{f.filename}` already exists and contains "
                f"information about {target} result. "
                f"If you wish to overwrite the file, set "
                f"`overwrite=True`."
            )


class ProblemHDF5Writer:
    """
    Writer of the HDF5 problem files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Initialize writer.

        Parameters
        ----------
        storage_filename: str
            HDF5 problem file name
        """
        self.storage_filename = storage_filename

    def write(self, problem, overwrite: bool = False):
        """Write HDF5 problem file from pyPESTO problem object."""
        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'problem')
            attrs_to_save = [
                a
                for a in dir(problem)
                if not a.startswith('__')
                and not callable(getattr(problem, a))
                and not hasattr(type(problem), a)
            ]

            problem_grp = f.create_group("problem")
            # save the configuration
            f['problem/config'] = str(problem.objective.get_config())

            for problem_attr in attrs_to_save:
                value = getattr(problem, problem_attr)
                if isinstance(value, (list, np.ndarray)):
                    value = np.asarray(value)
                    if value.size:
                        write_array(problem_grp, problem_attr, value)
                elif isinstance(value, Integral):
                    problem_grp.attrs[problem_attr] = value


class OptimizationResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite=False):
        """Write HDF5 result file from pyPESTO result object."""
        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'optimization')
            optimization_grp = f.require_group("optimization")
            # settings =
            # optimization_grp.create_dataset("settings", settings, dtype=)
            results_grp = optimization_grp.require_group("results")

            for start in result.optimize_result.list:
                start_id = start['id']
                start_grp = results_grp.require_group(start_id)
                for key in start.keys():
                    if key == 'history':
                        continue
                    if isinstance(start[key], np.ndarray):
                        write_float_array(start_grp, key, start[key])
                    elif start[key] is not None:
                        start_grp.attrs[key] = start[key]
                f.flush()


class SamplingResultHDF5Writer:
    """
    Writer of the HDF5 sampling files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite: bool = False):
        """Write HDF5 sampling file from pyPESTO result object."""
        # if there is no sample available, log a warning and return
        # SampleResult is only a dummy class created by the Result class
        # and always indicates the lack of a sampling result.
        if isinstance(result.sample_result, SampleResult):
            logger.warning(
                "Warning: There is no sampling_result, "
                "which you tried to save to hdf5."
            )
            return

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'sampling')
            results_grp = f.require_group("sampling/results")

            for key in result.sample_result.keys():
                if isinstance(result.sample_result[key], np.ndarray):
                    write_float_array(
                        results_grp, key, result.sample_result[key]
                    )
                elif result.sample_result[key] is not None:
                    results_grp.attrs[key] = result.sample_result[key]
            f.flush()


class ProfileResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite: bool = False):
        """Write HDF5 result file from pyPESTO result object."""
        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'profiling')
            profiling_grp = f.require_group("profiling")

            for profile_id, profile in enumerate(result.profile_result.list):
                profile_grp = profiling_grp.require_group(str(profile_id))
                for parameter_id, parameter_profile in enumerate(profile):
                    result_grp = profile_grp.require_group(str(parameter_id))

                    if parameter_profile is None:
                        result_grp.attrs['IsNone'] = True
                        continue
                    result_grp.attrs['IsNone'] = False
                    for key in parameter_profile.keys():
                        if isinstance(parameter_profile[key], np.ndarray):
                            write_float_array(
                                result_grp, key, parameter_profile[key]
                            )
                        elif parameter_profile[key] is not None:
                            result_grp.attrs[key] = parameter_profile[key]
            f.flush()


def write_result(
    result: Result,
    filename: str,
    overwrite: bool = False,
    problem: bool = True,
    optimize: bool = False,
    profile: bool = False,
    sample: bool = False,
):
    """
    Save whole pypesto.Result to hdf5 file.

    Boolean indicators allow specifying what to save.

    Parameters
    ----------
    result:
        The :class:`pypesto.Result` object to be saved.
    filename:
        The HDF5 filename.
    overwrite:
        Boolean, whether already existing results should be overwritten.
    problem:
        Read the problem.
    optimize:
        Read the optimize result.
    profile:
        Read the profile result.
    sample:
        Read the sample result.
    """
    if not any([optimize, profile, sample]):
        optimize = True
        profile = True
        sample = True

    if problem:
        pypesto_problem_writer = ProblemHDF5Writer(filename)
        pypesto_problem_writer.write(result.problem, overwrite=overwrite)

    if optimize:
        pypesto_opt_writer = OptimizationResultHDF5Writer(filename)
        pypesto_opt_writer.write(result, overwrite=overwrite)

    if profile:
        pypesto_profile_writer = ProfileResultHDF5Writer(filename)
        pypesto_profile_writer.write(result, overwrite=overwrite)

    if sample:
        pypesto_sample_writer = SamplingResultHDF5Writer(filename)
        pypesto_sample_writer.write(result, overwrite=overwrite)



"""Track optimal values during an optimization."""

import logging
from typing import Tuple, Union

import numpy as np

from ..C import FVAL, GRAD, HESS, RES, SRES, ModeType, X
from ..util import allclose, is_none_or_nan, is_none_or_nan_array, isclose
from .base import HistoryBase, add_fun_from_res
from .util import ResultDict

logger = logging.getLogger(__name__)


class OptimizerHistory:
    """
    Optimizer objective call history.

    Container around a History object, additionally keeping track of optimal
    values.

    Attributes
    ----------
    fval0, fval_min:
        Initial and best function value found.
    x0, x_min:
        Initial and best parameters found.
    grad_min:
        gradient for best parameters
    hess_min:
        hessian (approximation) for best parameters
    res_min:
        residuals for best parameters
    sres_min:
        residual sensitivities for best parameters

    Parameters
    ----------
    history:
        History object to attach to this container. This history object
        implements the storage of the actual history.
    x0:
        Initial values for optimization.
    lb, ub:
        Lower and upper bound. Used for checking validity of optimal points.
    generate_from_history:
        If set to true, this function will try to fill attributes of this
        function based on the provided history.
    """

    # optimal point values
    MIN_KEYS = (X, *HistoryBase.RESULT_KEYS)

    def __init__(
        self,
        history: HistoryBase,
        x0: np.ndarray,
        lb: np.ndarray,
        ub: np.ndarray,
        generate_from_history: bool = False,
    ) -> None:
        self.history: HistoryBase = history

        # initial point
        self.fval0: Union[float, None] = None
        self.x0: np.ndarray = x0

        # bounds
        self.lb: np.ndarray = lb
        self.ub: np.ndarray = ub

        # minimum point
        self.fval_min: float = np.inf
        self.x_min: Union[np.ndarray, None] = None
        self.grad_min: Union[np.ndarray, None] = None
        self.hess_min: Union[np.ndarray, None] = None
        self.res_min: Union[np.ndarray, None] = None
        self.sres_min: Union[np.ndarray, None] = None

        if generate_from_history:
            self._maybe_compute_init_and_min_vals_from_trace()

    def update(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """Update history and best found value."""
        result = add_fun_from_res(result)
        self._update_vals(x, result)
        self.history.update(x, sensi_orders, mode, result)

    def finalize(self, message: str = None, exitflag: int = None):
        """
        Finalize history.

        Parameters
        ----------
        message:
            Optimizer message to be saved.
        exitflag:
            Optimizer exitflag to be saved.
        """
        self.history.finalize(message=message, exitflag=exitflag)

        # There can be entries in the history e.g. for grad that are not
        #  recorded in ..._min, e.g. when evaluated before fval.
        # On the other hand, not all variables may be recorded in the history.
        # Thus, here at the end we go over the history once and try to fill
        #  in what is available.

        # check if a useful history exists
        # TODO Y This can be solved prettier
        try:
            self.history.get_x_trace()
        except NotImplementedError:
            return

        # find optimal point
        result = self._get_optimal_point_from_history()

        fval = result[FVAL]
        if fval is None:
            # nothing to be improved
            return

        # check if history has a better point (should not really happen)
        if (
            fval < self.fval_min
            and not isclose(fval, self.fval_min)
            and not allclose(result[X], self.x_min)
        ):
            # issue a warning, as if this happens, then something may be wrong
            logger.warn(
                f"History has a better point {fval} than the current best "
                "point {self.fval_min}."
            )
            # update everything
            for key in self.MIN_KEYS:
                setattr(self, key + '_min', result[key])

        # check if history has same point
        if (
            isclose(fval, self.fval_min)
            and self.x_min is not None
            and allclose(result[X], self.x_min)
        ):
            # update only missing entries
            #  (e.g. grad and hess may be recorded but not in history)
            for key in self.MIN_KEYS:
                if result[key] is not None:
                    # if getattr(self, f'{key}_min') is None:
                    setattr(self, f'{key}_min', result[key])

    def _update_vals(self, x: np.ndarray, result: ResultDict) -> None:
        """Update initial and best function values."""
        # update initial point
        if is_none_or_nan(self.fval0) and np.array_equal(x, self.x0):
            self.fval0 = result.get(FVAL)

        # don't update optimal point if point is not admissible
        if not self._admissible(x):
            return

        # update if fval is better
        if (
            not is_none_or_nan(fval := result.get(FVAL))
            and fval < self.fval_min
        ):
            # need to update all values, as better fval found
            for key in HistoryBase.RESULT_KEYS:
                setattr(self, f'{key}_min', result.get(key))
            self.x_min = x
            return

        # Sometimes sensitivities are evaluated on subsequent calls. We can
        # identify this situation by checking that x hasn't changed.
        if self.x_min is not None and np.array_equal(self.x_min, x):
            for key in (GRAD, HESS, SRES):
                val_min = getattr(self, f'{key}_min', None)
                if is_none_or_nan_array(val_min) and not is_none_or_nan_array(
                    val := result.get(key)
                ):
                    setattr(self, f'{key}_min', val)

    def _maybe_compute_init_and_min_vals_from_trace(self) -> None:
        """Try to set initial and best function value from trace.

        Only possible if history has a trace.
        """
        if not len(self.history):
            # nothing to be computed from empty history
            return

        # some optimizers may evaluate hess+grad first to compute trust region
        # etc
        for ix in range(len(self.history)):
            fval = self.history.get_fval_trace(ix)
            x = self.history.get_x_trace(ix)
            if not is_none_or_nan(fval) and allclose(x, self.x0):
                self.fval0 = float(fval)
                break

        # find best fval
        result = self._get_optimal_point_from_history()

        # assign values
        for key in OptimizerHistory.MIN_KEYS:
            setattr(self, f'{key}_min', result[key])

    def _admissible(self, x: np.ndarray) -> bool:
        """Check whether point `x` is admissible (i.e. within bounds).

        Parameters
        ----------
        x: A single parameter vector.

        Returns
        -------
        admissible: Whether the point fulfills the problem requirements.
        """
        return np.all(x <= self.ub) and np.all(x >= self.lb)

    def _get_optimal_point_from_history(self) -> ResultDict:
        """Extract optimal point from `self.history`."""
        result = {}

        # get indices of admissible trace entries
        # shape (n_sample, n_x)
        xs = np.asarray(self.history.get_x_trace())
        ixs_admit = [ix for ix, x in enumerate(xs) if self._admissible(x)]

        if len(ixs_admit) == 0:
            # no admittable indices
            return {key: None for key in OptimizerHistory.MIN_KEYS}

        # index of minimum of fval values
        ix_min = np.nanargmin(self.history.get_fval_trace(ixs_admit))
        # np.argmin returns ndarray when multiple minimal values are found,
        #  we want the first occurrence
        if isinstance(ix_min, np.ndarray):
            ix_min = ix_min[0]
        # select index in original array
        ix_min = ixs_admit[ix_min]

        # fill in parameter and function value from that index
        for var in (X, FVAL, RES):
            val = getattr(self.history, f'get_{var}_trace')(ix_min)
            if val is not None and not np.all(np.isnan(val)):
                result[var] = val
            # convert to float if var is FVAL to be sure
            if var == FVAL:
                result[var] = float(result[var])

        # derivatives may be evaluated at different indices, therefore
        #  iterate over all and check whether any has the same parameter
        #  and the desired field filled
        for var in (GRAD, HESS, SRES):
            for ix in range(len(self.history)):
                if not allclose(result[X], self.history.get_x_trace(ix)):
                    # different parameter
                    continue
                val = getattr(self.history, f'get_{var}_trace')(ix)
                if not is_none_or_nan_array(val):
                    result[var] = val
                    # successfuly found
                    break

        # fill remaining keys with None
        for key in OptimizerHistory.MIN_KEYS:
            if key not in result:
                result[key] = None

        return result



"""HDF5 history."""


import contextlib
import time
from typing import Dict, Sequence, Tuple, Union

import h5py
import numpy as np

from ..C import (
    EXITFLAG,
    FVAL,
    GRAD,
    HESS,
    HISTORY,
    MESSAGE,
    MESSAGES,
    MODE_FUN,
    MODE_RES,
    N_FVAL,
    N_GRAD,
    N_HESS,
    N_ITERATIONS,
    N_RES,
    N_SRES,
    RES,
    SRES,
    START_TIME,
    TIME,
    TRACE,
    TRACE_SAVE_ITER,
    ModeType,
    X,
)
from .base import HistoryBase, add_fun_from_res, reduce_result_via_options
from .options import HistoryOptions
from .util import MaybeArray, ResultDict, trace_wrap


def with_h5_file(mode: str):
    """Wrap function to work with hdf5 file.

    Parameters
    ----------
    mode:
        Access mode, see
        https://docs.h5py.org/en/stable/high/file.html.
    """
    modes = ["r", "a"]
    if mode not in modes:
        # can be extended if reasonable
        raise ValueError(f"Mode must be one of {modes}")

    def decorator(fun):
        def wrapper(self, *args, **kwargs):
            # file already opened
            if self._f is not None and (
                mode == self._f.mode
                or mode == "r"
                or (self._f.mode == "r+" and mode == "a")
            ):
                return fun(self, *args, **kwargs)

            with h5py.File(self.file, mode) as f:
                self._f = f
                ret = fun(self, *args, **kwargs)
                self._f = None
                return ret

        return wrapper

    return decorator


def check_editable(fun):
    """Check if the history is editable."""

    def wrapper(self, *args, **kwargs):
        if not self.editable:
            raise ValueError(
                f'ID "{self.id}" is already used in history file '
                f'"{self.file}".'
            )
        return fun(self, *args, **kwargs)

    return wrapper


class Hdf5History(HistoryBase):
    """
    Stores a representation of the history in an HDF5 file.

    Parameters
    ----------
    id:
        Id of the history
    file:
        HDF5 file name.
    options:
        History options.
    """

    def __init__(
        self,
        id: str,
        file: str,
        options: Union[HistoryOptions, Dict] = None,
    ):
        super().__init__(options=options)
        self.id: str = id
        self.file: str = file

        # filled during file access
        self._f: Union[h5py.File, None] = None

        # to check whether the trace can be edited
        self.editable: bool = self._editable()

    @check_editable
    @with_h5_file("a")
    def update(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """See `History` docstring."""
        # check whether the file was marked as editable upon initialization
        super().update(x, sensi_orders, mode, result)
        self._update_counts(sensi_orders, mode)
        self._update_trace(x, sensi_orders, mode, result)

    @with_h5_file("a")
    @check_editable
    def finalize(self, message: str = None, exitflag: str = None) -> None:
        """See `HistoryBase` docstring."""
        super().finalize()

        # add message and exitflag to trace
        f = self._f
        if f'{HISTORY}/{self.id}/{MESSAGES}/' not in f:
            f.create_group(f'{HISTORY}/{self.id}/{MESSAGES}/')
        grp = f[f'{HISTORY}/{self.id}/{MESSAGES}/']
        if message is not None:
            grp.attrs[MESSAGE] = message
        if exitflag is not None:
            grp.attrs[EXITFLAG] = exitflag

    @staticmethod
    def load(
        id: str, file: str, options: Union[HistoryOptions, Dict] = None
    ) -> 'Hdf5History':
        """Load the History object from memory."""
        history = Hdf5History(id=id, file=file, options=options)
        if options is None:
            history.recover_options(file)
        return history

    def recover_options(self, file: str):
        """Recover options when loading the hdf5 history from memory.

        Done by testing which entries were recorded.
        """
        trace_record = self._has_non_nan_entries(X)
        trace_record_grad = self._has_non_nan_entries(GRAD)
        trace_record_hess = self._has_non_nan_entries(HESS)
        trace_record_res = self._has_non_nan_entries(RES)
        trace_record_sres = self._has_non_nan_entries(SRES)

        restored_history_options = HistoryOptions(
            trace_record=trace_record,
            trace_record_grad=trace_record_grad,
            trace_record_hess=trace_record_hess,
            trace_record_res=trace_record_res,
            trace_record_sres=trace_record_sres,
            trace_save_iter=self.trace_save_iter,
            storage_file=file,
        )

        self.options = restored_history_options

    def _has_non_nan_entries(self, hdf5_group: str) -> bool:
        """Check if there exist non-nan entries stored for a given group."""
        group = self._get_hdf5_entries(hdf5_group, ix=None)

        for entry in group:
            if not (entry is None or np.all(np.isnan(entry))):
                return True

        return False

    @with_h5_file("a")
    def _update_counts(self, sensi_orders: Tuple[int, ...], mode: ModeType):
        """Update the counters in the hdf5."""
        group = self._require_group()

        if mode == MODE_FUN:
            if 0 in sensi_orders:
                group.attrs[N_FVAL] += 1
            if 1 in sensi_orders:
                group.attrs[N_GRAD] += 1
            if 2 in sensi_orders:
                group.attrs[N_HESS] += 1
        elif mode == MODE_RES:
            if 0 in sensi_orders:
                group.attrs[N_RES] += 1
            if 1 in sensi_orders:
                group.attrs[N_SRES] += 1

    @with_h5_file("r")
    def __len__(self) -> int:
        """Define length of history object."""
        try:
            return self._get_group().attrs[N_ITERATIONS]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def n_fval(self) -> int:
        """See `HistoryBase` docstring."""
        try:
            return self._get_group().attrs[N_FVAL]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def n_grad(self) -> int:
        """See `HistoryBase` docstring."""
        try:
            return self._get_group().attrs[N_GRAD]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def n_hess(self) -> int:
        """See `HistoryBase` docstring."""
        try:
            return self._get_group().attrs[N_HESS]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def n_res(self) -> int:
        """See `HistoryBase` docstring."""
        try:
            return self._get_group().attrs[N_RES]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def n_sres(self) -> int:
        """See `HistoryBase` docstring."""
        try:
            return self._get_group().attrs[N_SRES]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def trace_save_iter(self) -> int:
        """After how many iterations to store the trace."""
        try:
            return self._get_group().attrs[TRACE_SAVE_ITER]
        except KeyError:
            return 0

    @property
    @with_h5_file("r")
    def start_time(self) -> float:
        """See `HistoryBase` docstring."""
        # TODO Y This should also be saved in and recovered from the hdf5 file
        try:
            return self._get_group().attrs[START_TIME]
        except KeyError:
            return np.nan

    @property
    @with_h5_file("r")
    def message(self) -> str:
        """Optimizer message in case of finished optimization."""
        try:
            return self._f[f'{HISTORY}/{self.id}/{MESSAGES}/'].attrs[MESSAGE]
        except KeyError:
            return None

    @property
    @with_h5_file("r")
    def exitflag(self) -> str:
        """Optimizer exitflag in case of finished optimization."""
        try:
            return self._f[f'{HISTORY}/{self.id}/{MESSAGES}/'].attrs[EXITFLAG]
        except KeyError:
            return None

    @with_h5_file("a")
    def _update_trace(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """Update and possibly store the trace."""
        if not self.options.trace_record:
            return

        # calculating function values from residuals
        #  and reduce via requested history options
        result = reduce_result_via_options(
            add_fun_from_res(result), self.options
        )

        used_time = time.time() - self.start_time

        values = {
            X: x,
            FVAL: result[FVAL],
            GRAD: result[GRAD],
            RES: result[RES],
            SRES: result[SRES],
            HESS: result[HESS],
            TIME: used_time,
        }

        iteration = self._require_group().attrs[N_ITERATIONS]

        for key in values.keys():
            if values[key] is not None:
                self._require_group()[f'{iteration}/{key}'] = values[key]

        self._require_group().attrs[N_ITERATIONS] += 1

    @with_h5_file("r")
    def _get_group(self) -> h5py.Group:
        """Get the HDF5 group for the current history."""
        return self._f[f'{HISTORY}/{self.id}/{TRACE}/']

    @with_h5_file("a")
    def _require_group(self) -> h5py.Group:
        """Get, or if necessary create, the group in the hdf5 file."""
        with contextlib.suppress(KeyError):
            return self._f[f'{HISTORY}/{self.id}/{TRACE}/']

        grp = self._f.create_group(f'{HISTORY}/{self.id}/{TRACE}/')
        grp.attrs[N_ITERATIONS] = 0
        grp.attrs[N_FVAL] = 0
        grp.attrs[N_GRAD] = 0
        grp.attrs[N_HESS] = 0
        grp.attrs[N_RES] = 0
        grp.attrs[N_SRES] = 0
        grp.attrs[START_TIME] = time.time()
        # TODO Y it makes no sense to save this here
        #  Also, we do not seem to evaluate this at all
        grp.attrs[TRACE_SAVE_ITER] = self.options.trace_save_iter
        return grp

    @with_h5_file("r")
    def _get_hdf5_entries(
        self,
        entry_id: str,
        ix: Union[int, Sequence[int], None] = None,
    ) -> Sequence:
        """
        Get entries for field `entry_id` from HDF5 file, for indices `ix`.

        Parameters
        ----------
        entry_id:
            The key whose trace is returned.
        ix:
            Index or list of indices of the iterations that will produce
            the trace.

        Returns
        -------
        The entries ix for the key entry_id.
        """
        if ix is None:
            ix = range(len(self))
        trace_result = []

        for iteration in ix:
            try:
                dataset = self._f[
                    f'{HISTORY}/{self.id}/{TRACE}/{iteration}/{entry_id}'
                ]
                if dataset.shape == ():
                    entry = dataset[()]  # scalar
                else:
                    entry = np.array(dataset)
                trace_result.append(entry)
            except KeyError:
                trace_result.append(None)

        return trace_result

    @trace_wrap
    def get_x_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(X, ix)

    @trace_wrap
    def get_fval_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(FVAL, ix)

    @trace_wrap
    def get_grad_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(GRAD, ix)

    @trace_wrap
    def get_hess_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(HESS, ix)

    @trace_wrap
    def get_res_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(RES, ix)

    @trace_wrap
    def get_sres_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(SRES, ix)

    @trace_wrap
    def get_time_trace(
        self, ix: Union[int, Sequence[int], None] = None, trim: bool = False
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return self._get_hdf5_entries(TIME, ix)

    def _editable(self) -> bool:
        """
        Check whether the id is already existent in the file.

        Parameters
        ----------
        file:
            HDF5 file name.

        Returns
        -------
        True if the file is editable, False otherwise.
        """
        try:
            with h5py.File(self.file, "a") as f:
                # editable if the id entry does not exist
                if HISTORY not in f.keys() or self.id not in f[HISTORY]:
                    return True
                return False
        except OSError:
            # if something goes wrong, we assume the file is not editable
            return False



import logging
from typing import Callable

import pypesto.optimize

from ..engine import Task
from ..problem import Problem
from ..result import ProfilerResult
from .options import ProfileOptions
from .walk_along_profile import walk_along_profile

logger = logging.getLogger(__name__)


class ProfilerTask(Task):
    """A parameter likelihood profiling task."""

    def __init__(
        self,
        current_profile: ProfilerResult,
        problem: Problem,
        options: ProfileOptions,
        i_par: int,
        global_opt: float,
        optimizer: 'pypesto.optimize.Optimizer',
        create_next_guess: Callable,
    ):
        """
        Create the task object.

        Parameters
        ----------
        current_profile:
            The profile which should be computed
        problem:
            The problem to be solved.
        optimizer:
            The optimizer to be used along each profile.
        global_opt:
            log-posterior value of the global optimum
        options:
            Various options applied to the profile optimization.
        create_next_guess:
            Handle of the method which creates the next profile point proposal
        i_par:
            index for the current parameter
        """
        super().__init__()

        self.optimizer = optimizer
        self.problem = problem
        self.current_profile = current_profile
        self.global_opt = global_opt
        self.create_next_guess = create_next_guess
        self.i_par = i_par
        self.options = options

    def execute(self) -> 'pypesto.profile.ProfilerResult':
        """Compute profile in descending and ascending direction."""
        logger.debug(f"Executing task {self.i_par}.")

        for par_direction in [-1, 1]:
            # flip profile
            self.current_profile.flip_profile()

            # compute the current profile
            self.current_profile = walk_along_profile(
                current_profile=self.current_profile,
                problem=self.problem,
                par_direction=par_direction,
                optimizer=self.optimizer,
                options=self.options,
                create_next_guess=self.create_next_guess,
                global_opt=self.global_opt,
                i_par=self.i_par,
            )

        # return the ProfilerResult and the index of the parameter profiled
        return {'profile': self.current_profile, 'index': self.i_par}



from typing import Dict, Sequence, Union

import numpy as np
from tqdm import tqdm

from ..history import NoHistory
from ..objective import NegLogPriors, ObjectiveBase
from ..problem import Problem
from ..result import McmcPtResult
from .sampler import InternalSample, InternalSampler


class MetropolisSampler(InternalSampler):
    """Simple Metropolis-Hastings sampler with fixed proposal variance."""

    def __init__(self, options: Dict = None):
        super().__init__(options)
        self.problem: Union[Problem, None] = None
        self.neglogpost: Union[ObjectiveBase, None] = None
        self.neglogprior: Union[NegLogPriors, None] = None
        self.trace_x: Union[Sequence[np.ndarray], None] = None
        self.trace_neglogpost: Union[Sequence[float], None] = None
        self.trace_neglogprior: Union[Sequence[float], None] = None
        self.temper_lpost: bool = False

    @classmethod
    def default_options(cls):
        """Return the default options for the sampler."""
        return {
            'std': 1.0,  # the proposal standard deviation
            'show_progress': True,  # whether to show the progress
        }

    def initialize(self, problem: Problem, x0: np.ndarray):
        """Initialize the sampler."""
        self.problem = problem
        self.neglogpost = problem.objective
        self.neglogpost.history = NoHistory()
        if problem.x_priors is None:
            self.neglogprior = lambda x: -0.0
        else:
            self.neglogprior = problem.x_priors
        self.trace_x = [x0]
        self.trace_neglogpost = [self.neglogpost(x0)]
        self.trace_neglogprior = [self.neglogprior(x0)]

    def sample(self, n_samples: int, beta: float = 1.0):
        """Load last recorded particle."""
        x = self.trace_x[-1]
        lpost = -self.trace_neglogpost[-1]
        lprior = -self.trace_neglogprior[-1]

        show_progress = self.options['show_progress']

        # loop over iterations
        for _ in tqdm(range(int(n_samples)), disable=not show_progress):
            # perform step
            x, lpost, lprior = self._perform_step(
                x=x, lpost=lpost, lprior=lprior, beta=beta
            )
            # record step
            self.trace_x.append(x)
            self.trace_neglogpost.append(-lpost)
            self.trace_neglogprior.append(-lprior)

    def make_internal(self, temper_lpost: bool):
        """
        Allow the inner samplers to be used as inner samplers.

        Can be called by parallel tempering samplers during initialization.
        Default: Do nothing.

        Parameters
        ----------
        temper_lpost:
            Whether to temperate the posterior or only the likelihood.
        """
        self.options['show_progress'] = False
        self.temper_lpost = temper_lpost

    def _perform_step(
        self, x: np.ndarray, lpost: np.ndarray, lprior: np.ndarray, beta: float
    ):
        """
        Perform a step.

        Propose new parameter, evaluate and check whether to accept.
        """
        # propose step
        x_new: np.ndarray = self._propose_parameter(x)

        # check if step lies within bounds
        if any(x_new < self.problem.lb) or any(x_new > self.problem.ub):
            # will not be accepted
            lpost_new = -np.inf
        else:
            # compute log posterior
            lpost_new = -self.neglogpost(x_new)

        # check posterior evaluation is successful
        if np.isnan(lpost_new):
            # will not be accepted
            lpost_new = -np.inf

        # compute log prior
        lprior_new = -self.neglogprior(x_new)

        if not self.temper_lpost:
            # extract current log likelihood value
            llh = lpost - lprior
            # extract proposed log likelihood value
            llh_new = lpost_new - lprior_new
            # log acceptance probability (temper log likelihood)
            log_p_acc = min(beta * (llh_new - llh) + (lprior_new - lprior), 0)
        else:
            # log acceptance probability (temper log posterior)
            log_p_acc = min(beta * (lpost_new - lpost), 0)

        # flip a coin
        u = np.random.uniform(0, 1)

        # check acceptance
        if np.log(u) < log_p_acc:
            # update particle
            x = x_new
            lpost = lpost_new
            lprior = lprior_new

        # update proposal
        self._update_proposal(
            x, lpost, log_p_acc, len(self.trace_neglogpost) + 1
        )

        return x, lpost, lprior

    def _propose_parameter(self, x: np.ndarray):
        """Propose a step."""
        x_new: np.ndarray = x + self.options['std'] * np.random.randn(len(x))
        return x_new

    def _update_proposal(
        self, x: np.ndarray, lpost: float, log_p_acc: float, n_sample_cur: int
    ):
        """Update the proposal density. Default: Do nothing."""

    def get_last_sample(self) -> InternalSample:
        """Get the last sample in the chain.

        Returns
        -------
        internal_sample:
            The last sample in the chain in the exchange format.
        """
        return InternalSample(
            x=self.trace_x[-1],
            lpost=-self.trace_neglogpost[-1],
            lprior=-self.trace_neglogprior[-1],
        )

    def set_last_sample(self, sample: InternalSample):
        """
        Set the last sample in the chain to the passed value.

        Parameters
        ----------
        sample:
            The sample that will replace the last sample in the chain.
        """
        self.trace_x[-1] = sample.x
        self.trace_neglogpost[-1] = -sample.lpost
        self.trace_neglogprior[-1] = -sample.lprior

    def get_samples(self) -> McmcPtResult:
        """Get the samples into the fitting pypesto format."""
        result = McmcPtResult(
            trace_x=np.array([self.trace_x]),
            trace_neglogpost=np.array([self.trace_neglogpost]),
            trace_neglogprior=np.array([self.trace_neglogprior]),
            betas=np.array([1.0]),
        )
        return result



import abc
import copy
import os
import tempfile
from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, Optional, Sequence, Tuple, Union

import numpy as np

from ...C import FVAL, MODE_FUN, MODE_RES, RDATAS, ModeType
from ..base import ObjectiveBase, ResultDict
from .amici_calculator import AmiciCalculator
from .amici_util import (
    create_identity_parameter_mapping,
    map_par_opt_to_par_sim,
)

if TYPE_CHECKING:
    try:
        import amici
        from amici.parameter_mapping import ParameterMapping
    except ImportError:
        pass

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']


class AmiciObjectBuilder(abc.ABC):
    """Allows to build AMICI model, solver, and edatas.

    This class is useful for pickling an :class:`pypesto.AmiciObjective`,
    which is required in some parallelization schemes. Therefore, this
    class itself must be picklable.
    """

    @abc.abstractmethod
    def create_model(self) -> AmiciModel:
        """Create an AMICI model."""

    @abc.abstractmethod
    def create_solver(self, model: AmiciModel) -> AmiciSolver:
        """Create an AMICI solver."""

    @abc.abstractmethod
    def create_edatas(self, model: AmiciModel) -> Sequence['amici.ExpData']:
        """Create AMICI experimental data."""


class AmiciObjective(ObjectiveBase):
    """Allows to create an objective directly from an amici model."""

    def __init__(
        self,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: Union[Sequence['amici.ExpData'], 'amici.ExpData'],
        max_sensi_order: Optional[int] = None,
        x_ids: Optional[Sequence[str]] = None,
        x_names: Optional[Sequence[str]] = None,
        parameter_mapping: Optional['ParameterMapping'] = None,
        guess_steadystate: Optional[Optional[bool]] = None,
        n_threads: Optional[int] = 1,
        fim_for_hess: Optional[bool] = True,
        amici_object_builder: Optional[AmiciObjectBuilder] = None,
        calculator: Optional[AmiciCalculator] = None,
        amici_reporting: Optional['amici.RDataReporting'] = None,
    ):
        """
        Initialize objective.

        Parameters
        ----------
        amici_model:
            The amici model.
        amici_solver:
            The solver to use for the numeric integration of the model.
        edatas:
            The experimental data. If a list is passed, its entries correspond
            to multiple experimental conditions.
        max_sensi_order:
            Maximum sensitivity order supported by the model. Defaults to 2 if
            the model was compiled with o2mode, otherwise 1.
        x_ids:
            Ids of optimization parameters. In the simplest case, this will be
            the AMICI model parameters (default).
        x_names:
            Names of optimization parameters.
        parameter_mapping:
            Mapping of optimization parameters to model parameters. Format
            as created by `amici.petab_objective.create_parameter_mapping`.
            The default is just to assume that optimization and simulation
            parameters coincide.
        guess_steadystate:
            Whether to guess steadystates based on previous steadystates and
            respective derivatives. This option may lead to unexpected
            results for models with conservation laws and should accordingly
            be deactivated for those models.
        n_threads:
            Number of threads that are used for parallelization over
            experimental conditions. If amici was not installed with openMP
            support this option will have no effect.
        fim_for_hess:
            Whether to use the FIM whenever the Hessian is requested. This only
            applies with forward sensitivities.
            With adjoint sensitivities, the true Hessian will be used,
            if available.
            FIM or Hessian will only be exposed if `max_sensi_order>1`.
        amici_object_builder:
            AMICI object builder. Allows recreating the objective for
            pickling, required in some parallelization schemes.
        calculator:
            Performs the actual calculation of the function values and
            derivatives.
        amici_reporting:
            Determines which quantities will be computed by AMICI,
            see ``amici.Solver.setReturnDataReportingMode``. Set to ``None``
            to compute only the minimum required information.
        """
        import amici

        if amici is None:
            raise ImportError(
                "This objective requires an installation of amici "
                "(https://github.com/icb-dcm/amici). "
                "Install via `pip3 install amici`."
            )

        self.amici_model = amici_model.clone()
        self.amici_solver = amici_solver.clone()

        # make sure the edatas are a list of edata objects
        if isinstance(edatas, amici.amici.ExpData):
            edatas = [edatas]

        # set the experimental data container
        self.edatas = edatas

        # set the maximum sensitivity order
        self.max_sensi_order = max_sensi_order

        self.guess_steadystate = guess_steadystate

        # optimization parameter ids
        if x_ids is None:
            # use model parameter ids as ids
            x_ids = list(self.amici_model.getParameterIds())
        self.x_ids = x_ids

        # mapping of parameters
        if parameter_mapping is None:
            # use identity mapping for each condition
            parameter_mapping = create_identity_parameter_mapping(
                amici_model, len(edatas)
            )
        self.parameter_mapping = parameter_mapping

        # If supported, enable `guess_steadystate` by default. If not
        #  supported, disable by default. If requested but unsupported, raise.
        if (
            self.guess_steadystate is not False
            and self.amici_model.nx_solver_reinit > 0
        ):
            if self.guess_steadystate:
                raise ValueError(
                    'Steadystate prediction is not supported '
                    'for models with conservation laws!'
                )
            self.guess_steadystate = False

        if (
            self.guess_steadystate is not False
            and self.amici_model.getSteadyStateSensitivityMode()
            == amici.SteadyStateSensitivityMode.integrationOnly
        ):
            if self.guess_steadystate:
                raise ValueError(
                    'Steadystate guesses cannot be enabled '
                    'when `integrationOnly` as '
                    'SteadyStateSensitivityMode!'
                )
            self.guess_steadystate = False

        if self.guess_steadystate is not False:
            self.guess_steadystate = True

        if self.guess_steadystate:
            # preallocate guesses, construct a dict for every edata for which
            #  we need to do preequilibration
            self.steadystate_guesses = {
                'fval': np.inf,
                'data': {
                    iexp: {}
                    for iexp, edata in enumerate(self.edatas)
                    if len(edata.fixedParametersPreequilibration)
                },
            }
        # optimization parameter names
        if x_names is None:
            # use ids as names
            x_names = x_ids

        self.n_threads = n_threads
        self.fim_for_hess = fim_for_hess
        self.amici_object_builder = amici_object_builder
        self.amici_reporting = amici_reporting

        if calculator is None:
            calculator = AmiciCalculator()
        self.calculator = calculator
        super().__init__(x_names=x_names)

        # Custom (condition-specific) timepoints. See the
        # `set_custom_timepoints` method for more information.
        self.custom_timepoints = None

    def get_config(self) -> dict:
        """Return basic information of the objective configuration."""
        info = super().get_config()
        info['x_names'] = self.x_names
        info['model_name'] = self.amici_model.getName()
        info['solver'] = str(type(self.amici_solver))
        info['sensi_order'] = self.max_sensi_order

        return info

    def initialize(self):
        """See `ObjectiveBase` documentation."""
        super().initialize()
        self.reset_steadystate_guesses()
        self.calculator.initialize()

    def __deepcopy__(self, memodict: Dict = None) -> 'AmiciObjective':
        import amici

        other = self.__class__.__new__(self.__class__)

        for key in set(self.__dict__.keys()) - {
            'amici_model',
            'amici_solver',
            'edatas',
        }:
            other.__dict__[key] = copy.deepcopy(self.__dict__[key])

        # copy objects that do not have __deepcopy__
        other.amici_model = self.amici_model.clone()
        other.amici_solver = self.amici_solver.clone()
        other.edatas = [amici.ExpData(data) for data in self.edatas]

        return other

    def __getstate__(self) -> Dict:
        import amici

        if self.amici_object_builder is None:
            raise NotImplementedError(
                "AmiciObjective does not support __getstate__ without "
                "an `amici_object_builder`."
            )

        state = {}
        for key in set(self.__dict__.keys()) - {
            'amici_model',
            'amici_solver',
            'edatas',
        }:
            state[key] = self.__dict__[key]

        _fd, _file = tempfile.mkstemp()
        try:
            # write amici solver settings to file
            try:
                amici.writeSolverSettingsToHDF5(self.amici_solver, _file)
            except AttributeError as e:
                e.args += (
                    "Pickling the AmiciObjective requires an AMICI "
                    "installation with HDF5 support.",
                )
                raise
            # read in byte stream
            with open(_fd, 'rb', closefd=False) as f:
                state['amici_solver_settings'] = f.read()
        finally:
            # close file descriptor and remove temporary file
            os.close(_fd)
            os.remove(_file)

        state['AMICI_model_settings'] = amici.get_model_settings(
            self.amici_model
        )

        return state

    def __setstate__(self, state: Dict) -> None:
        import amici

        if state['amici_object_builder'] is None:
            raise NotImplementedError(
                "AmiciObjective does not support __setstate__ without "
                "an `amici_object_builder`."
            )
        self.__dict__.update(state)

        # note: attributes not defined in the builder are lost
        model = self.amici_object_builder.create_model()
        solver = self.amici_object_builder.create_solver(model)
        edatas = self.amici_object_builder.create_edatas(model)

        _fd, _file = tempfile.mkstemp()
        try:
            # write solver settings to temporary file
            with open(_fd, 'wb', closefd=False) as f:
                f.write(state['amici_solver_settings'])
            # read in solver settings
            try:
                amici.readSolverSettingsFromHDF5(_file, solver)
            except AttributeError as err:
                if not err.args:
                    err.args = ('',)
                err.args += (
                    "Unpickling an AmiciObjective requires an AMICI "
                    "installation with HDF5 support.",
                )
                raise
        finally:
            # close file descriptor and remove temporary file
            os.close(_fd)
            os.remove(_file)

        self.amici_model = model
        self.amici_solver = solver
        self.edatas = edatas

        self.apply_custom_timepoints()
        amici.set_model_settings(
            self.amici_model,
            state['AMICI_model_settings'],
        )

    def check_sensi_orders(
        self,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
    ) -> bool:
        """See `ObjectiveBase` documentation."""
        import amici

        if not sensi_orders:
            return True
        sensi_order = max(sensi_orders)

        # dynamically obtain maximum allowed sensitivity order
        max_sensi_order = self.max_sensi_order
        if max_sensi_order is None:
            max_sensi_order = 1
            # check whether it is ok to request 2nd order
            sensi_mthd = self.amici_solver.getSensitivityMethod()
            mthd_fwd = amici.SensitivityMethod_forward
            if mode == MODE_FUN and (
                self.amici_model.o2mode
                or (sensi_mthd == mthd_fwd and self.fim_for_hess)
            ):
                max_sensi_order = 2

        # evaluate sensitivity order
        return sensi_order <= max_sensi_order

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return mode in [MODE_FUN, MODE_RES]

    def __call__(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...] = (0,),
        mode: ModeType = MODE_FUN,
        return_dict: bool = False,
        **kwargs,
    ) -> Union[float, np.ndarray, Tuple, ResultDict]:
        """See `ObjectiveBase` documentation."""
        import amici

        # Use AMICI full reporting if amici.ReturnDatas are returned and no
        #  other reporting mode was set
        if (
            return_dict
            and self.amici_reporting is None
            and 'amici_reporting' not in kwargs
        ):
            kwargs['amici_reporting'] = amici.RDataReporting.full

        return super().__call__(x, sensi_orders, mode, return_dict, **kwargs)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        edatas: Sequence['amici.ExpData'] = None,
        parameter_mapping: 'ParameterMapping' = None,
        amici_reporting: Optional['amici.RDataReporting'] = None,
    ):
        """
        Call objective function without pre- or post-processing and formatting.

        Returns
        -------
        result:
            A dict containing the results.
        """
        import amici

        x_dct = self.par_arr_to_dct(x)

        # only ask amici to compute required quantities
        amici_reporting = (
            self.amici_reporting
            if amici_reporting is None
            else amici_reporting
        )
        if amici_reporting is None:
            amici_reporting = (
                amici.RDataReporting.likelihood
                if mode == MODE_FUN
                else amici.RDataReporting.residuals
            )
        self.amici_solver.setReturnDataReportingMode(amici_reporting)

        # update steady state
        if (
            self.guess_steadystate
            and self.steadystate_guesses['fval'] < np.inf
        ):
            for data_ix in range(len(self.edatas)):
                self.apply_steadystate_guess(data_ix, x_dct)

        if edatas is None:
            edatas = self.edatas
        if parameter_mapping is None:
            parameter_mapping = self.parameter_mapping
        ret = self.calculator(
            x_dct=x_dct,
            sensi_orders=sensi_orders,
            mode=mode,
            amici_model=self.amici_model,
            amici_solver=self.amici_solver,
            edatas=edatas,
            n_threads=self.n_threads,
            x_ids=self.x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=self.fim_for_hess,
        )

        nllh = ret[FVAL]
        rdatas = ret[RDATAS]

        # check whether we should update data for preequilibration guesses
        if (
            self.guess_steadystate
            and nllh <= self.steadystate_guesses['fval']
            and nllh < np.inf
        ):
            self.steadystate_guesses['fval'] = nllh
            for data_ix, rdata in enumerate(rdatas):
                self.store_steadystate_guess(data_ix, x_dct, rdata)

        return ret

    def par_arr_to_dct(self, x: Sequence[float]) -> Dict[str, float]:
        """Create dict from parameter vector."""
        return OrderedDict(zip(self.x_ids, x))

    def apply_steadystate_guess(self, condition_ix: int, x_dct: Dict) -> None:
        """
        Apply steady state guess to `edatas[condition_ix].x0`.

        Use the stored steadystate as well as the respective sensitivity (
        if available) and parameter value to approximate the steadystate at
        the current parameters using a zeroth or first order taylor
        approximation:
        x_ss(x') = x_ss(x) [+ dx_ss/dx(x)*(x'-x)]
        """
        mapping = self.parameter_mapping[condition_ix].map_sim_var
        x_sim = map_par_opt_to_par_sim(mapping, x_dct, self.amici_model)
        x_ss_guess = []  # resets initial state by default
        if condition_ix in self.steadystate_guesses['data']:
            guess_data = self.steadystate_guesses['data'][condition_ix]
            if guess_data['x_ss'] is not None:
                x_ss_guess = guess_data['x_ss']
            if guess_data['sx_ss'] is not None:
                linear_update = (
                    guess_data['sx_ss']
                    .transpose()
                    .dot(
                        (x_sim - guess_data['x'])[
                            np.asarray(self.edatas[condition_ix].plist)
                        ]
                    )
                )
                # limit linear updates to max 20 % elementwise change
                if (linear_update / (x_ss_guess + np.spacing(1))).max() < 0.2:
                    x_ss_guess += linear_update

        self.edatas[condition_ix].x0 = tuple(x_ss_guess)

    def store_steadystate_guess(
        self,
        condition_ix: int,
        x_dct: Dict,
        rdata: 'amici.ReturnData',
    ) -> None:
        """
        Store condition parameter, steadystate and steadystate sensitivity.

        Stored in steadystate_guesses if steadystate guesses are enabled for
        this condition.
        """
        if condition_ix not in self.steadystate_guesses['data']:
            return
        preeq_guesses = self.steadystate_guesses['data'][condition_ix]

        # update parameter
        condition_map_sim_var = self.parameter_mapping[
            condition_ix
        ].map_sim_var
        x_sim = map_par_opt_to_par_sim(
            condition_map_sim_var, x_dct, self.amici_model
        )
        preeq_guesses['x'] = x_sim

        # update steadystates
        preeq_guesses['x_ss'] = rdata['x_ss']
        preeq_guesses['sx_ss'] = rdata['sx_ss']

    def reset_steadystate_guesses(self) -> None:
        """Reset all steadystate guess data."""
        if not self.guess_steadystate:
            return

        self.steadystate_guesses['fval'] = np.inf
        for condition in self.steadystate_guesses['data']:
            self.steadystate_guesses['data'][condition] = {}

    def apply_custom_timepoints(self) -> None:
        """Apply custom timepoints, if applicable.

        See the `set_custom_timepoints` method for more information.
        """
        if self.custom_timepoints is not None:
            for index in range(len(self.edatas)):
                self.edatas[index].setTimepoints(self.custom_timepoints[index])

    def set_custom_timepoints(
        self,
        timepoints: Sequence[Sequence[Union[float, int]]] = None,
        timepoints_global: Sequence[Union[float, int]] = None,
    ) -> 'AmiciObjective':
        """
        Create a copy of this objective that is evaluated at custom timepoints.

        The intended use is to aid in predictions at unmeasured timepoints.

        Parameters
        ----------
        timepoints:
            The outer sequence should contain a sequence of timepoints for each
            experimental condition.
        timepoints_global:
            A sequence of timepoints that will be used for all experimental
            conditions.

        Returns
        -------
        The customized copy of this objective.
        """
        if timepoints is None and timepoints_global is None:
            raise KeyError('Timepoints were not specified.')

        amici_objective = copy.deepcopy(self)

        if timepoints is not None:
            if len(timepoints) != len(amici_objective.edatas):
                raise ValueError(
                    'The number of condition-specific timepoints `timepoints` '
                    'does not match the number of experimental conditions.\n'
                    f'Number of provided timepoints: {len(timepoints)}. '
                    'Number of experimental conditions: '
                    f'{len(amici_objective.edatas)}.'
                )
            custom_timepoints = timepoints
        else:
            custom_timepoints = [
                copy.deepcopy(timepoints_global)
                for _ in range(len(amici_objective.edatas))
            ]

        amici_objective.custom_timepoints = custom_timepoints
        amici_objective.apply_custom_timepoints()
        return amici_objective

    def check_gradients_match_finite_differences(
        self, x: np.ndarray = None, *args, **kwargs
    ) -> bool:
        """Check if gradients match finite differences (FDs).

        Parameters
        ----------
        x: The parameters for which to evaluate the gradient.

        Returns
        -------
        bool
            Indicates whether gradients match (True) FDs or not (False)
        """
        if x is None and 'petab_problem' in dir(self.amici_object_builder):
            x = self.amici_object_builder.petab_problem.x_nominal_scaled
            x_free = self.amici_object_builder.petab_problem.x_free_indices
        return super().check_gradients_match_finite_differences(
            x=x, x_free=x_free, *args, **kwargs
        )



"""Cooperative Enhanced Scatter Search."""
import logging
import multiprocessing
import os
import time
from typing import Dict, List, Optional

import numpy as np

import pypesto.optimize
from pypesto import Problem
from pypesto.startpoint import StartpointMethod

from .ess import ESSExitFlag, ESSOptimizer
from .function_evaluator import FunctionEvaluator
from .refset import RefSet

logger = logging.getLogger(__name__)


class CESSOptimizer:
    r"""
    Cooperative Enhanced Scatter Search Optimizer (CESS).

    A cooperative scatter search algorithm based on [VillaverdeEge2012]_.
    In short, multiple scatter search instances with different hyperparameters
    are running in different threads/processes, and exchange information.
    Some instances focus on diversification while others focus on
    intensification. Communication happens at fixed time intervals.

    Proposed hyperparameter values in [VillaverdeEge2012]_:

    * ``dim_refset``: ``[0.5 n_parameter, 20 n_parameters]``
    * ``local_n2``: ``[0, 100]``
    * ``balance``: ``[0, 0.5]``
    * ``n_diverse``: ``[5 n_par, 20 n_par]``
    * ``max_eval``: such that :math:`\tau = log10(max_eval / n_par)` is in
      [2.5, 3.5], with a recommended default value of 2.5.

    .. [VillaverdeEge2012] 'A cooperative strategy for parameter estimation in
       large scale systems biology models', Villaverde, A.F., Egea,
       J.A. & Banga, J.R. BMC Syst Biol 2012, 6, 75.
       https://doi.org/10.1186/1752-0509-6-75

    Attributes
    ----------
    ess_init_args:
        List of argument dictionaries passed to
        :func:`ESSOptimizer.__init__`. The length of this list is the
        number of parallel ESS processes.
        Resource limits such as ``max_eval`` apply to a single CESS
        iteration, not to the full search.
    max_iter:
        Maximum number of CESS iterations.
    max_walltime_s:
        Maximum walltime in seconds. Will only be checked between local
        optimizations and other simulations, and thus, may be exceeded by
        the duration of a local search. Defaults to no limit.
    fx_best:
        The best objective value seen so far.
    x_best:
        Parameter vector corresponding to ``fx_best``.
    starttime:
        Starting time of the most recent optimization.
    i_iter:
        Current iteration number.
    """

    def __init__(
        self,
        ess_init_args: List[Dict],
        max_iter: int,
        max_walltime_s: float = np.inf,
    ):
        """Construct.

        Parameters
        ----------
        ess_init_args:
            List of argument dictionaries passed to
            :func:`ESSOptimizer.__init__`. The length of this list is the
            number of parallel ESS processes.
            Resource limits such as ``max_eval`` apply to a single CESS
            iteration, not to the full search.
        max_iter:
            Maximum number of CESS iterations.
        max_walltime_s:
            Maximum walltime in seconds. Will only be checked between local
            optimizations and other simulations, and thus, may be exceeded by
            the duration of a local search. Defaults to no limit.
        """
        self.max_walltime_s = max_walltime_s
        self.ess_init_args = ess_init_args
        self.max_iter = max_iter

        self._initialize()

    def _initialize(self):
        """(Re-)initialize."""
        self.starttime = time.time()
        self.i_iter = 0
        # Overall best parameters found so far
        self.x_best: Optional[np.array] = None
        # Overall best function value found so far
        self.fx_best: float = np.inf

    def minimize(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
    ) -> pypesto.Result:
        """Minimize the given objective using CESS.

        Parameters
        ----------
        problem:
            Problem to run ESS on.
        startpoint_method:
            Method for choosing starting points.
        """
        self._initialize()

        refsets = [None] * len(self.ess_init_args)

        evaluator = FunctionEvaluator(
            problem=problem,
            startpoint_method=startpoint_method,
            n_threads=1,
        )

        while True:
            logger.info("-" * 50)
            self._report_iteration()
            self.i_iter += 1

            # run scatter searches
            ess_optimizers = self._run_scatter_searches(
                problem=problem,
                startpoint_method=startpoint_method,
                refsets=refsets,
            )
            # collect refsets from the different ESS runs
            refsets = [result.refset for result in ess_optimizers]

            # update best values from ESS results
            for result in ess_optimizers:
                self._maybe_update_global_best(result.x_best, result.fx_best)

            if not self._keep_going(i_eval=evaluator.n_eval):
                break

            # create refsets for the next iteration
            self._update_refsets(refsets=refsets, evaluator=evaluator)

            # TODO merge results
        self._report_final()

        # TODO what should the result look like?
        return self._create_result(problem, refsets)

    def _report_iteration(self):
        """Log the current iteration."""
        if self.max_iter == 0:
            logger.info("iter | best |")

        with np.printoptions(
            edgeitems=30,
            linewidth=100000,
            formatter={"float": lambda x: "%.3g" % x},
        ):
            logger.info(f"{self.i_iter:4} | {self.fx_best:+.2E} | ")

    def _report_final(self):
        """Log scatter search summary."""
        with np.printoptions(
            edgeitems=30,
            linewidth=100000,
            formatter={"float": lambda x: "%.3g" % x},
        ):
            logger.info(
                f"CESS finished with {self.exit_flag!r} "
                f"after {self.i_iter} iterations, "
                f"{time.time() - self.starttime:.3g}s. "
                # f"Num local solutions: {len(self.local_solutions)}."
            )
            # logger.info(f"Final refset: {np.sort(self.refset.fx)} ")
        logger.info(f"Best fval {self.fx_best}")

    def _create_result(
        self, problem: pypesto.Problem, refsets: List[RefSet]
    ) -> pypesto.Result:
        """Create the result object.

        Currently, this returns the overall best value and the final RefSet.
        """
        common_result_fields = {
            'exitflag': self.exit_flag,
            # meaningful? this is the overall time, and identical for all
            #  reported points
            'time': time.time() - self.starttime,
            # TODO
            # 'n_fval': self.evaluator.n_eval,
            'optimizer': str(self),
        }
        i_result = 0
        result = pypesto.Result(problem=problem)

        # save global best
        optimizer_result = pypesto.OptimizerResult(
            id=str(i_result),
            x=self.x_best,
            fval=self.fx_best,
            message="Global best",
            **common_result_fields,
        )
        # TODO DW: Create a single History with the global best?
        result.optimize_result.append(optimizer_result)

        # save refsets
        for i_refset, refset in enumerate(refsets):
            for i in range(refset.dim):
                i_result += 1
                result.optimize_result.append(
                    pypesto.OptimizerResult(
                        id=str(i_result),
                        x=refset.x[i],
                        fval=refset.fx[i],
                        message=f"RefSet[{i_refset}][{i}]",
                        **common_result_fields,
                    )
                )

        # TODO DW: also save local solutions?
        #  (need to track fvals or re-evaluate)

        return result

    def _run_scatter_searches(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
        refsets: List[RefSet],
    ) -> List[ESSOptimizer]:
        """Start all scatter searches in different processes."""
        # set default value of max_eval if not present.
        #  only set it on a copy, as the original dicts may be re-used
        #  for different optimization problems.
        # reasonable value proposed in [VillaverdeEge2012]:
        #  2.5 < tau < 3.5, default: 2.5
        ess_init_args = [
            dict(
                {'max_eval': int(10**2.5 * problem.dim)},
                **ess_kwargs,
            )
            for ess_kwargs in self.ess_init_args
        ]

        ctx = multiprocessing.get_context('spawn')
        with ctx.Pool(len(self.ess_init_args)) as pool:
            ess_optimizers = pool.starmap(
                self._run_single_ess,
                (
                    [ess_kwargs, problem, startpoint_method, refset]
                    for (ess_kwargs, refset) in zip(ess_init_args, refsets)
                ),
                chunksize=1,
            )
        return list(ess_optimizers)

    def _run_single_ess(
        self,
        ess_kwargs,
        problem: Problem,
        startpoint_method: StartpointMethod,
        refset: RefSet,
    ) -> ESSOptimizer:
        """
        Run ESS.

        Helper for `starmap`.
        """
        # different random seeds per process?
        np.random.seed((os.getpid() * int(time.time() * 1000)) % 2**32)

        ess = ESSOptimizer(**ess_kwargs)
        ess.minimize(problem, startpoint_method, refset=refset)
        return ess

    def _keep_going(self, i_eval) -> bool:
        """Check exit criteria.

        Returns
        -------
        ``True`` if none of the exit criteria is met, ``False`` otherwise.
        """
        # TODO DW which further stopping criteria: gtol, fatol, frtol?

        # elapsed iterations
        if self.i_iter >= self.max_iter:
            self.exit_flag = ESSExitFlag.MAX_ITER
            return False

        # elapsed time
        if time.time() - self.starttime >= self.max_walltime_s:
            self.exit_flag = ESSExitFlag.MAX_TIME
            return False

        return True

    def _maybe_update_global_best(self, x, fx):
        """Update the global best value if the provided value is better."""
        if fx < self.fx_best:
            self.x_best = x[:]
            self.fx_best = fx

    def _update_refsets(
        self, refsets: List[RefSet], evaluator: FunctionEvaluator
    ):
        """
        Update refsets.

        Create new refsets based on the combined final refsets of the previous
        CESS iteration.

        Updates ``refsets`` in place.
        """
        # gather final refset entries
        x = np.vstack([refset.x for refset in refsets])
        fx = np.concatenate([refset.fx for refset in refsets])

        # reset function evaluation counter
        evaluator.n_eval = 0
        evaluator.n_eval_round = 0

        for i, ess_init_args in enumerate(self.ess_init_args):
            refsets[i] = RefSet(
                dim=ess_init_args['dim_refset'], evaluator=evaluator
            )
            refsets[i].initialize_from_array(x_diverse=x, fx_diverse=fx)
            refsets[i].sort()



from typing import Dict, Sequence

import numpy as np

from ..C import GRAD, HESS, RES, SRES


class PrePostProcessor:
    """
    Implements the methods preprocess and postprocess.

    They are called at the beginning and at the end of the objective call,
    in order to handle the mapping of optimization parameters to simulation
    parameters.

    This class acts as a dummy base implementation, not performing any
    changes on the passed objects.
    """

    def __init__(self):
        pass

    def preprocess(self, x: np.ndarray) -> np.ndarray:  # pylint: disable=R0201
        """
        Just return x without modifications.

        Parameters
        ----------
        x:
            Parameter vector for optimization.

        Returns
        -------
        x:
            Parameter vector for simulation.
        """
        return x

    def postprocess(self, result: Dict) -> Dict:  # pylint: disable=R0201
        """
        Convert all arrays into np.ndarrays if necessary, and return them.

        Parameters
        ----------
        result:
            The result object to finalize.
        """
        result = PrePostProcessor.as_ndarrays(result)
        return result

    def reduce(self, x: np.ndarray) -> np.ndarray:  # pylint: disable=R0201
        """
        Return x without modifications.

        Parameters
        ----------
        x:
            Parameter vector for simulation.

        Returns
        -------
        x:
            Parameter vector for optimization.
        """
        return x

    @staticmethod
    def as_ndarrays(result: Dict) -> Dict:
        """
        Convert all array_like objects to np.ndarrays.

        This has the advantage of a uniform output datatype which offers
        various methods to assess the data.
        """
        keys = [GRAD, HESS, RES, SRES]
        for key in keys:
            if key in result:
                value = result[key]
                if value is not None:
                    result[key] = np.array(value)
        return result


class FixedParametersProcessor(PrePostProcessor):
    """Extends the processor to handle the fixing of parameters."""

    def __init__(
        self,
        dim_full: int,
        x_free_indices: Sequence[int],
        x_fixed_indices: Sequence[int],
        x_fixed_vals: Sequence[float],
    ):
        super().__init__()
        self.dim_full: int = dim_full
        self.x_free_indices: np.ndarray = np.array(x_free_indices, dtype=int)
        self.x_fixed_indices: np.ndarray = np.array(x_fixed_indices, dtype=int)
        self.x_fixed_vals: np.ndarray = np.array(x_fixed_vals, dtype=float)

    def preprocess(self, x: np.ndarray) -> np.ndarray:
        """Embed optimization vector to full vector with all parameters."""
        x = super().preprocess(x)

        x_full = np.zeros(self.dim_full)
        x_full[self.x_free_indices] = x
        x_full[self.x_fixed_indices] = self.x_fixed_vals

        return x_full

    def reduce(self, x: np.ndarray) -> np.ndarray:
        """
        Return x reduced to free indices.

        Parameters
        ----------
        x:
            Parameter vector for simulation.

        Returns
        -------
        x:
            Parameter vector for optimization.
        """
        x = super().reduce(x)

        if x.size:
            return x[self.x_free_indices]
        else:
            return x

    def postprocess(self, result: Dict) -> Dict:
        """Constrain results to optimization parameter dimensions."""
        result = super().postprocess(result)

        if result.get(GRAD, None) is not None:
            grad = result[GRAD]
            if grad.size == self.dim_full:
                grad = grad[self.x_free_indices]
                result[GRAD] = grad
        if result.get(HESS, None) is not None:
            hess = result[HESS]
            if hess.shape[0] == self.dim_full:
                hess = hess[np.ix_(self.x_free_indices, self.x_free_indices)]
                result[HESS] = hess
        if result.get(SRES, None) is not None:
            sres = result[SRES]
            if sres.shape[-1] == self.dim_full:
                sres = sres[..., self.x_free_indices]
                result[SRES] = sres

        return result



"""ReferenceSet functionality for scatter search."""

from typing import Optional

import numpy as np

from .function_evaluator import FunctionEvaluator


class RefSet:
    """Scatter search reference set.

    Attributes
    ----------
    dim:
        Reference set size
    evaluator:
        Function evaluator
    x:
        Parameters in the reference set
    fx:
        Function values at the parameters in the reference set
    n_stuck:
        Counts the number of times a refset member did not lead to an
        improvement in the objective (length: ``dim``).
    """

    def __init__(
        self,
        dim: int,
        evaluator: FunctionEvaluator,
        x: Optional[np.array] = None,
        fx: Optional[np.array] = None,
    ):
        """Construct.

        Parameters
        ----------
        dim:
            Reference set size
        evaluator:
            Function evaluator
        x:
            Initial RefSet parameters.
        fx:
            Function values corresponding to entries in x. Must be provided if
            and only if ``x`` is not ``None``.
        """
        if (x is not None and fx is None) or (x is None and fx is not None):
            raise ValueError(
                "Either both or neither of `x` and `fx` " "should be provided"
            )

        if dim < 3:
            raise ValueError("RefSet dimension has to be at least 3.")
        self.dim = dim
        self.evaluator = evaluator
        # \epsilon in [PenasGon2017]_
        self.proximity_threshold = 1e-3

        if x is None:
            self.x = self.fx = None
        else:
            self.x = x
            self.fx = fx

        self.n_stuck = np.zeros(shape=[dim])

    def sort(self):
        """Sort RefSet by quality."""
        order = np.argsort(self.fx)
        self.fx = self.fx[order]
        self.x = self.x[order]
        self.n_stuck = self.n_stuck[order]

    def initialize_random(
        self,
        n_diverse: int = None,
    ):
        """Create initial reference set from random parameters.

        Sample ``n_diverse`` random points, populate half of the RefSet using
        the best solutions and fill the rest with random points.
        """
        # sample n_diverse points
        x_diverse, fx_diverse = self.evaluator.multiple_random(n_diverse)

        self.initialize_from_array(x_diverse=x_diverse, fx_diverse=fx_diverse)

    def initialize_from_array(self, x_diverse: np.array, fx_diverse: np.array):
        """Create initial reference set using the provided points.

        Populate half of the RefSet using the best given solutions and fill the
        rest with a random selection from the remaining points.
        """
        if len(x_diverse) != len(fx_diverse):
            raise ValueError(
                "Lengths of `x_diverse` and `fx_diverse` do " "not match."
            )
        if self.dim > len(x_diverse):
            raise ValueError(
                "Cannot create RefSet with dimension "
                f"{self.dim} from only {len(x_diverse)} points."
            )

        self.fx = np.full(shape=(self.dim,), fill_value=np.inf)
        self.x = np.full(
            shape=(self.dim, self.evaluator.problem.dim), fill_value=np.nan
        )

        # create initial refset with 50% best values
        num_best = int(self.dim / 2)
        order = np.argsort(fx_diverse)
        self.x[:num_best] = x_diverse[order[:num_best]]
        self.fx[:num_best] = fx_diverse[order[:num_best]]

        # ... and 50% random points
        random_idxs = np.random.choice(
            order[num_best:], size=self.dim - num_best, replace=False
        )
        self.x[num_best:] = x_diverse[random_idxs]
        self.fx[num_best:] = fx_diverse[random_idxs]

    def prune_too_close(self):
        """Prune too similar RefSet members.

        Replace a parameter vector if its maximum relative difference to a
        better parameter vector is below the given threshold.

        Assumes RefSet is sorted.
        """
        x = self.x
        for i in range(self.dim):
            for j in range(i + 1, self.dim):
                # check proximity
                # zero-division may occur here
                with np.errstate(divide='ignore', invalid='ignore'):
                    while (
                        np.max(np.abs((x[i] - x[j]) / x[j]))
                        <= self.proximity_threshold
                    ):
                        # too close. replace x_j.
                        x[j], self.fx[j] = self.evaluator.single_random()
                        self.sort()

    def update(self, i: int, x: np.array, fx: float):
        """Update a RefSet entry."""
        self.x[i] = x
        self.fx[i] = fx
        self.n_stuck[i] = 0

    def replace_by_random(self, i: int):
        """Replace the RefSet member with the given index by a random point."""
        self.x[i], self.fx[i] = self.evaluator.single_random()
        self.n_stuck[i] = 0



"""Pymc v4 Sampler."""
from __future__ import annotations

import logging
from typing import Union

import arviz as az
import numpy as np
import pymc
import pytensor.tensor as pt

from ..history import MemoryHistory
from ..objective import ObjectiveBase
from ..problem import Problem
from ..result import McmcPtResult
from .sampler import Sampler, SamplerImportError

logger = logging.getLogger(__name__)

# implementation based on:
# https://www.pymc.io/projects/examples/en/latest/case_studies/blackbox_external_likelihood_numpy.html


class PymcObjectiveOp(pt.Op):
    """PyTensor wrapper around a (non-normalized) log-probability function."""

    itypes = [pt.dvector]  # expects a vector of parameter values when called
    otypes = [pt.dscalar]  # outputs a single scalar value (the log prob)

    def create_instance(objective: ObjectiveBase, beta: float = 1.0):
        """Create an instance of this Op (factory method).

        Parameters
        ----------
        objective:
            Objective function (negative log-likelihood or -posterior).
        beta:
            Inverse temperature (e.g. in parallel tempering).

        Returns
        -------
        PymcObjectiveOp
            The created instance.
        """
        if objective.has_grad:
            return PymcObjectiveWithGradientOp(objective, beta)
        return PymcObjectiveOp(objective, beta)

    def __init__(self, objective: ObjectiveBase, beta: float = 1.0):
        self._objective: ObjectiveBase = objective
        self._beta: float = beta

    def perform(self, node, inputs, outputs, params=None):
        """Calculate the objective function value."""
        (theta,) = inputs
        log_prob = -self._beta * self._objective(theta, sensi_orders=(0,))
        outputs[0][0] = np.array(log_prob)


class PymcObjectiveWithGradientOp(PymcObjectiveOp):
    """PyTensor objective wrapper with gradient."""

    def __init__(self, objective: ObjectiveBase, beta: float = 1.0):
        super().__init__(objective, beta)

        self._log_prob_grad = PymcGradientOp(objective, beta)

    def grad(self, inputs, g):  # noqa
        """Calculate the vector-Jacobian product."""
        # the method that calculates the gradients - it actually returns the
        # vector-Jacobian product - g[0] is a vector of parameter values
        (theta,) = inputs  # our parameters
        return [g[0] * self._log_prob_grad(theta)]


class PymcGradientOp(pt.Op):
    """PyTensor wrapper around a (non-normalized) log-probability gradient."""

    itypes = [pt.dvector]  # expects a vector of parameter values when called
    otypes = [pt.dvector]  # outputs a vector (the log prob grad)

    def __init__(self, objective: ObjectiveBase, beta: float):
        self._objective: ObjectiveBase = objective
        self._beta: float = beta

    def perform(self, node, inputs, outputs, params=None):
        """Calculate the gradients of the objective function."""
        (theta,) = inputs
        # calculate gradients
        log_prob_grad = -self._beta * self._objective(theta, sensi_orders=(1,))
        outputs[0][0] = log_prob_grad


class PymcSampler(Sampler):
    """Wrapper around Pymc v4 samplers.

    Parameters
    ----------
    step_function:
        A pymc step function, e.g. NUTS, Slice. If not specified, pymc
        determines one automatically (preferable).
    **kwargs:
        Options are directly passed on to `pymc.sample`.
    """

    def __init__(
        self,
        step_function=None,
        post_compute_fval: bool = True,
        **kwargs,
    ):
        super().__init__(kwargs)
        self.step_function = step_function
        self.problem: Union[Problem, None] = None
        self.x0: Union[np.ndarray, None] = None
        self.trace: Union[pymc.backends.Text, None] = None
        self.data: Union[az.InferenceData, None] = None

    @classmethod
    def translate_options(cls, options):
        """
        Translate options and fill in defaults.

        Parameters
        ----------
        options:
            Options configuring the sampler.
        """
        if not options:
            options = {'chains': 1}
        return options

    def initialize(self, problem: Problem, x0: np.ndarray):
        """
        Initialize the sampler.

        Parameters
        ----------
        problem:
            The problem for which to sample.
        x0:
            Should, but is not required to, be used as initial parameter.
        """
        self.problem = problem
        if x0 is not None:
            if len(x0) != problem.dim:
                x0 = problem.get_reduced_vector(x0)
        self.x0 = x0
        self.trace = None
        self.data = None

        self.problem.objective.history = MemoryHistory()

    def sample(self, n_samples: int, beta: float = 1.0):
        """
        Sample the problem.

        Parameters
        ----------
        n_samples:
            Number of samples to be computed.
        """
        try:
            import pymc
        except ImportError:
            raise SamplerImportError("pymc")

        problem = self.problem
        log_post = PymcObjectiveOp.create_instance(problem.objective, beta)
        trace = self.trace

        x0 = None
        x_names_free = problem.get_reduced_vector(problem.x_names)
        if self.x0 is not None:
            x0 = {
                x_name: val
                for x_name, val in zip(problem.x_names, self.x0)
                if x_name in x_names_free
            }

        # create model context
        with pymc.Model():
            # parameter bounds as uniform prior
            _k = [
                pymc.Uniform(x_name, lower=lb, upper=ub)
                for x_name, lb, ub in zip(
                    x_names_free,
                    problem.lb,
                    problem.ub,
                )
            ]

            # convert parameters to PyTensor tensor variable
            theta = pt.as_tensor_variable(_k)

            # define distribution with log-posterior as density
            pymc.Potential("potential", log_post(theta))

            # record function values
            pymc.Deterministic("loggyposty", log_post(theta))

            # step, by default automatically determined by pymc
            step = None
            if self.step_function:
                step = self.step_function()

            # perform the actual sampling
            data = pymc.sample(
                draws=int(n_samples),
                trace=trace,
                initvals=x0,
                step=step,
                **self.options,
            )

        self.data = data

    def get_samples(self) -> McmcPtResult:
        """Convert result from pymc to McmcPtResult."""
        # dimensions
        n_par, n_chain, n_iter = np.asarray(
            self.data.posterior.to_array()
        ).shape
        n_par -= 1  # remove log-posterior

        # parameters
        trace_x = np.empty(shape=(n_chain, n_iter, n_par))
        par_ids = self.problem.get_reduced_vector(self.problem.x_names)
        if len(par_ids) != n_par:
            raise AssertionError("Mismatch of parameter dimension")
        for i_par, par_id in enumerate(par_ids):
            trace_x[:, :, i_par] = np.asarray(self.data.posterior[par_id])

        # function values
        trace_neglogpost = -np.asarray(self.data.posterior["loggyposty"])

        if (
            trace_x.shape[0] != trace_neglogpost.shape[0]
            or trace_x.shape[1] != trace_neglogpost.shape[1]
            or trace_x.shape[2] != self.problem.dim
        ):
            raise ValueError(
                "Trace dimensions are inconsistent: "
                f"{trace_x.shape=} {trace_neglogpost.shape=} {self.problem.dim=}"
            )

        return McmcPtResult(
            trace_x=np.array(trace_x),
            trace_neglogpost=np.array(trace_neglogpost),
            trace_neglogprior=np.full(trace_neglogpost.shape, np.nan),
            betas=np.array([1.0] * trace_x.shape[0]),
        )



import warnings
from numbers import Number
from typing import Iterable, List, Optional, Union

import numpy as np

from ..C import (
    ALL,
    ALL_CLUSTERED,
    FIRST_CLUSTER,
    FREE_ONLY,
    LEN_RGB,
    LEN_RGBA,
    RGB,
    RGB_RGBA,
    RGBA_ALPHA,
    RGBA_MAX,
    RGBA_MIN,
    RGBA_WHITE,
)
from ..result import Result
from ..util import assign_clusters, delete_nan_inf
from .clust_color import assign_colors_for_list


def process_result_list(
    results: Union[Result, List[Result]], colors=None, legends=None
):
    """
    Assign colors and legends to a list of results, check user provided lists.

    Parameters
    ----------
    results: list or pypesto.Result
        list of pypesto.Result objects or a single pypesto.Result
    colors: list, optional
        list of RGBA colors
    legends: str or list
        labels for line plots

    Returns
    -------
    results: list of pypesto.Result
       list of pypesto.Result objects
    colors: list of RGBA
        One for each element in 'results'.
    legends: list of str
        labels for line plots
    """
    # check how many results were passed
    single_result = False
    legend_error = False
    if isinstance(results, list):
        if len(results) == 1:
            single_result = True
    else:
        single_result = True
        results = [results]

    # handle results according to their number
    if single_result:
        # assign colors and create list for later handling
        if colors is None:
            colors = [colors]
        else:
            colors = [np.array(colors)]

        # create list of legends for later handling
        if not isinstance(legends, list):
            legends = [legends]
    else:
        # if more than one result is passed, we use one color per result
        colors = assign_colors_for_list(len(results), colors)

        # check whether list of legends has the correct length
        if legends is None:
            # No legends were passed: create some custom legends
            legends = []
            for i_leg in range(len(results)):
                legends.append('Result ' + str(i_leg))
        else:
            # legends were passed by user: check length
            if isinstance(legends, list):
                if len(legends) != len(results):
                    legend_error = True
            else:
                legend_error = True

    # size of legend list and size of results does not match
    if legend_error:
        raise ValueError(
            'List of results passed and list of labels do '
            'not have the same length but should. Stopping.'
        )

    return results, colors, legends


def process_offset_y(
    offset_y: Optional[float], scale_y: str, min_val: float
) -> float:
    """
    Compute offset for y-axis, depend on user settings.

    Parameters
    ----------
    offset_y:
       value for offsetting the later plotted values, in order to ensure
       positivity if a semilog-plot is used
    scale_y:
       Can be 'lin' or 'log10', specifying whether values should be plotted
       on linear or on log10-scale
    min_val:
        Smallest value to be plotted

    Returns
    -------
    offset_y: float
       value for offsetting the later plotted values, in order to ensure
       positivity if a semilog-plot is used
    """
    # check whether the offset specified by the user is sufficient
    if offset_y is not None:
        if (scale_y == 'log10') and (min_val + offset_y <= 0.0):
            warnings.warn(
                "Offset specified by user is insufficient. "
                "Ignoring specified offset and using "
                + str(np.abs(min_val) + 1.0)
                + " instead."
            )
        else:
            return offset_y
    else:
        # check whether scaling is lin or log10
        if scale_y == 'lin':
            # linear scaling doesn't need any offset
            return 0.0

    return 1.0 - min_val


def process_y_limits(ax, y_limits):
    """
    Apply user specified limits of y-axis.

    Parameters
    ----------
    ax: matplotlib.Axes, optional
        Axes object to use.
    y_limits: ndarray
       y_limits, minimum and maximum, for current axes object

    Returns
    -------
    ax: matplotlib.Axes, optional
        Axes object to use.
    """
    # apply y-limits, if they were specified by the user
    if y_limits is not None:
        y_limits = np.array(y_limits)

        # check validity of bounds
        if y_limits.size == 0:
            y_limits = np.array(ax.get_ylim())
        elif y_limits.size == 1:
            # if the user specified only an upper bound
            tmp_y_limits = ax.get_ylim()
            y_limits = [tmp_y_limits[0], y_limits]
        elif y_limits.size > 1:
            y_limits = [y_limits[0], y_limits[1]]

        # check validity of bounds if plotting in log-scale
        if ax.get_yscale() == 'log' and y_limits[0] <= 0.0:
            tmp_y_limits = ax.get_ylim()
            if y_limits[1] <= 0.0:
                y_limits = tmp_y_limits
                warnings.warn(
                    "Invalid bounds for plotting in "
                    "log-scale. Using defaults bounds."
                )
            else:
                y_limits = [tmp_y_limits[0], y_limits[1]]
                warnings.warn(
                    "Invalid lower bound for plotting in "
                    "log-scale. Using only upper bound."
                )

            # set limits
            ax.set_ylim(y_limits)

    else:
        # No limits passed, but if we have a result list: check the limits
        ax_limits = np.array(ax.get_ylim())
        data_limits = ax.dataLim.ymin, ax.dataLim.ymax

        # Check if data fits to axes and adapt limits, if necessary
        if ax_limits[0] > data_limits[0] or ax_limits[1] < data_limits[1]:
            # Get range of data
            data_range = data_limits[1] - data_limits[0]
            if ax.get_yscale() == 'log':
                data_range = np.log10(data_range)
                new_limits = (
                    np.power(10, np.log10(data_limits[0]) - 0.02 * data_range),
                    np.power(10, np.log10(data_limits[1]) + 0.02 * data_range),
                )
            else:
                new_limits = (
                    data_limits[0] - 0.02 * data_range,
                    data_limits[1] + 0.02 * data_range,
                )

            # set limits
            ax.set_ylim(new_limits)

    return ax


def rgba2rgb(fg: RGB_RGBA, bg: RGB_RGBA = None) -> RGB:
    """Combine two colors, removing transparency.

    Parameters
    ----------
    fg:
        Foreground color.
    bg:
        Background color.

    Returns
    -------
    The combined color.
    """
    if bg is None:
        bg = RGBA_WHITE
    if len(bg) == LEN_RGBA:
        # return foreground if background is fully transparent
        if bg[RGBA_ALPHA] == RGBA_MIN:
            return fg
    else:
        if len(bg) != LEN_RGB:
            raise IndexError(
                'A background color of unexpected length was provided: {bg}'
            )
        bg = (*bg, RGBA_MAX)

    # return the foreground color if has no transparency
    if len(fg) == LEN_RGB or fg[RGBA_ALPHA] == RGBA_MAX:
        return fg
    if len(fg) != LEN_RGBA:
        raise IndexError(
            'A foreground color of unexpected length was provided: {fg}'
        )

    def apparent_composite_color_component(
        fg_component: float,
        bg_component: float,
        fg_alpha: float = fg[RGBA_ALPHA],
        bg_alpha: float = bg[RGBA_ALPHA],
    ) -> float:
        """
        Composite a foreground over a background color component.

        Porter and Duff equations are used for alpha compositing.

        Parameters
        ----------
        fg_component:
            The foreground color component.
        bg_component:
            The background color component.
        fg_alpha:
            The foreground color transparency/alpha component.
        bg_alpha:
            The background color transparency/alpha component.

        Returns
        -------
        The component of the new color.
        """
        return (
            fg_component * fg_alpha
            + bg_component * bg_alpha * (RGBA_MAX - fg_alpha)
        ) / (fg_alpha + bg_alpha * (RGBA_MAX - fg_alpha))

    return [
        apparent_composite_color_component(fg[i], bg[i])
        for i in range(LEN_RGB)
    ]


def process_start_indices(
    result: Result,
    start_indices: Union[str, int, Iterable[int]] = None,
) -> np.ndarray:
    """
    Process the start_indices.

    Create an array of indices if a number was provided and checks that the
    indices do not exceed the max_index.

    Parameters
    ----------
    start_indices:
        list of indices or int specifying an endpoint of the sequence of
        indices. Furthermore the following strings are possible:
            * 'all', this is the default, using all start indices.
            * 'all_clustered', this includes the best start and all clusters
            with the size > 1.
            * 'first_cluster', includes all starts that belong to the first
            cluster.
    result:
        Result to determine maximum allowed length and/or clusters.
    """
    if start_indices is None:
        start_indices = ALL
    if isinstance(start_indices, str):
        if start_indices == ALL:
            return np.asarray(range(len(result.optimize_result)))
        elif start_indices == ALL_CLUSTERED:
            clust_ind, clust_size = assign_clusters(
                delete_nan_inf(result.optimize_result.fval)[1]
            )
            # get all clusters that have size >= 2 and cluster of best start:
            clust_gr2 = np.where(clust_size > 2)[0]
            clust_gr2 = (
                np.append(clust_gr2, 0) if 0 not in clust_gr2 else clust_gr2
            )
            start_indices = np.concatenate(
                [np.where(clust_ind == i_clust)[0] for i_clust in clust_gr2]
            )
            return start_indices
        elif start_indices == FIRST_CLUSTER:
            clust_ind = assign_clusters(
                delete_nan_inf(result.optimize_result.fval)[1]
            )[0]
            return np.where(clust_ind == 0)[0]
        else:
            raise ValueError(
                f"Permissible values for start_indices are {ALL}, "
                f"{ALL_CLUSTERED}, {FIRST_CLUSTER}, an integer or a "
                f"list of indices."
            )
    # if it is an integer n, select the first n starts
    if isinstance(start_indices, Number):
        start_indices = range(int(start_indices))

    # filter out the indices that exceed the range of possible start indices
    start_indices = [
        start_index
        for start_index in start_indices
        if start_index < len(result.optimize_result)
    ]

    return np.asarray(start_indices)


def process_parameter_indices(
    result: Result,
    parameter_indices: Union[str, Iterable[int]] = FREE_ONLY,
) -> list:
    """
    Process the parameter indices, always returning a valid array.

    Create an array of indices depending on the string that is provided. Or
    returns the sequence in case a sequence was provided.

    Parameters
    ----------
    result:
        Result to determine maximum allowed length and/or clusters.
    parameter_indices:
        list of indices or str specifying the desired indices. Default is
        `free_only`. Other option is 'all', which included all estimated
        and fixed parameters.
    """
    if isinstance(parameter_indices, str):
        if parameter_indices == ALL:
            return list(range(0, result.problem.dim_full))
        elif parameter_indices == FREE_ONLY:
            return result.problem.x_free_indices
        else:
            raise ValueError(
                "Permissible values for parameter_indices are "
                f"{ALL}, {FREE_ONLY} or a list of indices."
            )
    return list(parameter_indices)



"""History options."""

from pathlib import Path
from typing import Dict, Union

from ..C import SUFFIXES, SUFFIXES_CSV
from .util import CsvHistoryTemplateError, HistoryTypeError


class HistoryOptions(dict):
    """
    Options for what values to record.

    In addition implements a factory pattern to generate history objects.

    Parameters
    ----------
    trace_record:
        Flag indicating whether to record the trace of function calls.
        The trace_record_* flags only become effective if
        trace_record is True.
    trace_record_grad:
        Flag indicating whether to record the gradient in the trace.
    trace_record_hess:
        Flag indicating whether to record the Hessian in the trace.
    trace_record_res:
        Flag indicating whether to record the residual in
        the trace.
    trace_record_sres:
        Flag indicating whether to record the residual sensitivities in
        the trace.
    trace_save_iter:
        After how many iterations to store the trace.
    storage_file:
        File to save the history to. Can be any of None, a
        "{filename}.csv", or a "{filename}.hdf5" file. Depending on the values,
        the `create_history` method creates the appropriate object.
        Occurrences of "{id}" in the file name are replaced by the `id`
        upon creation of a history, if applicable.
    """

    def __init__(
        self,
        trace_record: bool = False,
        trace_record_grad: bool = True,
        trace_record_hess: bool = True,
        trace_record_res: bool = True,
        trace_record_sres: bool = True,
        trace_save_iter: int = 10,
        storage_file: str = None,
    ):
        super().__init__()

        self.trace_record: bool = trace_record
        self.trace_record_grad: bool = trace_record_grad
        self.trace_record_hess: bool = trace_record_hess
        self.trace_record_res: bool = trace_record_res
        self.trace_record_sres: bool = trace_record_sres
        self.trace_save_iter: int = trace_save_iter
        self.storage_file: str = storage_file

        self._sanity_check()

    def __getattr__(self, key):
        """Allow to use keys as attributes."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def _sanity_check(self):
        """Apply basic sanity checks."""
        if self.storage_file is None:
            return

        # extract storage type
        suffix = Path(self.storage_file).suffix[1:]

        # check storage format is valid
        if suffix not in SUFFIXES:
            raise HistoryTypeError(suffix)

        # check csv histories are parametrized
        if suffix in SUFFIXES_CSV and "{id}" not in self.storage_file:
            raise CsvHistoryTemplateError(self.storage_file)

    @staticmethod
    def assert_instance(
        maybe_options: Union['HistoryOptions', Dict],
    ) -> 'HistoryOptions':
        """
        Return a valid options object.

        Parameters
        ----------
        maybe_options: HistoryOptions or dict
        """
        if isinstance(maybe_options, HistoryOptions):
            return maybe_options
        options = HistoryOptions(**maybe_options)
        return options



from typing import List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import inset_locator

from pypesto.util import delete_nan_inf

from ..result import Result
from .clust_color import RGBA, assign_colors
from .misc import (
    process_offset_y,
    process_result_list,
    process_start_indices,
    process_y_limits,
)
from .reference_points import ReferencePoint, create_references


def waterfall(
    results: Union[Result, Sequence[Result]],
    ax: Optional[plt.Axes] = None,
    size: Optional[Tuple[float]] = (18.5, 10.5),
    y_limits: Optional[Tuple[float]] = None,
    scale_y: Optional[str] = 'log10',
    offset_y: Optional[float] = None,
    start_indices: Optional[Union[Sequence[int], int]] = None,
    n_starts_to_zoom: int = 0,
    reference: Optional[Sequence[ReferencePoint]] = None,
    colors: Optional[Union[RGBA, Sequence[RGBA]]] = None,
    legends: Optional[Union[Sequence[str], str]] = None,
    order_by_id: bool = False,
):
    """
    Plot waterfall plot.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    ax: matplotlib.Axes, optional
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    y_limits: float or ndarray, optional
        Maximum value to be plotted on the y-axis, or y-limits
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        Offset for the y-axis, if it is supposed to be in log10-scale
    start_indices:
        Integers specifying the multistart to be plotted or int specifying
        up to which start index should be plotted
    n_starts_to_zoom:
        Number of best multistarts that should be zoomed in.
        Should be smaller that the total number of multistarts
    reference:
        Reference points for optimization results, containing at least a
        function value fval
    colors:
        Colors or single color  for plotting. If not set, clustering is done
        and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    order_by_id:
        Function values corresponding to the same start ID will be located at
        the same x-axis position. Only applicable when a list of result
        objects are provided. Default behavior is to sort the function values
        of each result independently of other results.

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    if n_starts_to_zoom:
        # create zoom in
        inset_axes = inset_locator.inset_axes(
            ax, width="30%", height="30%", loc='center right'
        )
        inset_locator.mark_inset(ax, inset_axes, loc1=2, loc2=4)
    else:
        inset_axes = None

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    # handle `order_by_id`
    if order_by_id:
        start_id_ordering = get_ordering_by_start_id(results)

    refs = create_references(references=reference)

    # precompute y-offset, if needed and if a list of results was passed
    fvals_all, offset_y = process_offset_for_list(
        offset_y, results, scale_y, start_indices, refs
    )

    # plotting routine needs the maximum number of multistarts
    max_len_fvals = 0

    # loop over results
    for j, fvals_raw in enumerate(fvals_all):
        # extract specific cost function values from result
        max_len_fvals = np.max([max_len_fvals, *fvals_raw.shape])

        # remove colors where value is infinite if colors were passed on
        if colors[j] is not None and fvals_raw.size == colors[j].shape[0]:
            colors[j] = colors[j][
                np.isfinite(np.transpose(fvals_raw)).flatten()
            ]

        # parse input
        if order_by_id:
            start_ids = [s.id for s in results[j].optimize_result.list]
            fvals_raw_is_finite = np.isfinite(fvals_raw)

            fvals = []
            for start_id in start_id_ordering:
                start_index = start_ids.index(start_id)
                if fvals_raw_is_finite[start_index]:
                    fvals.append(fvals_raw[start_index])
                else:
                    fvals.append(None)
        else:
            # remove nan or inf values in fvals
            _, fvals = delete_nan_inf(fvals_raw)
            fvals.sort()

        # assign colors
        coloring = assign_colors(fvals, colors=colors[j])

        # call lowlevel plot routine
        ax = waterfall_lowlevel(
            fvals=fvals,
            scale_y=scale_y,
            offset_y=offset_y,
            ax=ax,
            size=size,
            colors=coloring,
            legend_text=legends[j],
        )

        if inset_axes is not None:
            inset_axes = waterfall_lowlevel(
                fvals=fvals[:n_starts_to_zoom],
                scale_y=scale_y,
                ax=inset_axes,
                colors=coloring[:n_starts_to_zoom],
            )

    # apply changes specified be the user to the axis object
    ax = handle_options(ax, max_len_fvals, refs, y_limits, offset_y)
    if inset_axes is not None:
        inset_axes = handle_options(
            inset_axes, n_starts_to_zoom, refs, y_limits, offset_y
        )

    if any(legends):
        ax.legend()
    # labels
    ax.set_xlabel('Ordered optimizer run')
    if offset_y == 0.0:
        ax.set_ylabel('Function value')
    else:
        ax.set_ylabel(f'Objective value (offset={offset_y:0.3e})')
    ax.set_title('Waterfall plot')
    return ax


def waterfall_lowlevel(
    fvals,
    ax: Optional[plt.Axes] = None,
    size: Optional[Tuple[float]] = (18.5, 10.5),
    scale_y: str = 'log10',
    offset_y: float = 0.0,
    colors: Optional[Union[RGBA, Sequence[RGBA]]] = None,
    legend_text: Optional[str] = None,
):
    """
    Plot waterfall plot using list of function values.

    Parameters
    ----------
    fvals: numeric list or array
        Including values need to be plotted. `None` values indicate that the
        corresponding start index should be skipped.
    ax: matplotlib.Axes
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    scale_y: str, optional
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        offset for the y-axis, if it is supposed to be in log10-scale
    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically
    legend_text:
        Label for line plots

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    start_indices = [i for i, fval in enumerate(fvals) if fval is not None]
    fvals = [fvals[i] for i in start_indices]

    # assign colors
    colors = assign_colors(fvals, colors=colors)

    # plot
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    # plot line
    if scale_y == 'log10':
        ax.semilogy(start_indices, fvals, color=[0.7, 0.7, 0.7, 0.6])
    else:
        ax.plot(start_indices, fvals, color=[0.7, 0.7, 0.7, 0.6])

    # plot points
    for j in start_indices:
        # parse data for plotting
        color = colors[j]
        fval = fvals[j]
        if j == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # line plot (linear or logarithmic)
        if scale_y == 'log10':
            ax.semilogy(
                j, fval, color=color, marker='o', label=tmp_legend, alpha=1.0
            )
        else:
            ax.plot(
                j, fval, color=color, marker='o', label=tmp_legend, alpha=1.0
            )

    # check if y-axis has a reasonable scale
    y_min, y_max = ax.get_ylim()
    if scale_y == 'log10':
        if np.log10(y_max) - np.log10(y_min) < 1.0:
            ax.set_ylim(
                ax.dataLim.y0 - 0.001 * abs(ax.dataLim.y0),
                ax.dataLim.y1 + 0.001 * abs(ax.dataLim.y1),
            )
    else:
        if y_max - y_min < 1.0:
            y_mean = 0.5 * (y_min + y_max)
            ax.set_ylim(y_mean - 0.5, y_mean + 0.5)

    # labels
    ax.set_xlabel('Ordered optimizer run')
    if offset_y == 0.0:
        ax.set_ylabel('Function value')
    else:
        ax.set_ylabel('Objective value (offset={offset_y:0.3e})')
    ax.set_title('Waterfall plot')
    if legend_text is not None:
        ax.legend()

    return ax


def process_offset_for_list(
    offset_y: float,
    results: Sequence[Result],
    scale_y: Optional[str],
    start_indices: Optional[Sequence[int]] = None,
    references: Optional[Sequence[ReferencePoint]] = None,
) -> Tuple[List[np.ndarray], float]:
    """
    Compute common offset_y and add it to `fvals` of results.

    Parameters
    ----------
    offset_y:
        User provided offset_y
    results:
        Optimization results obtained by 'optimize.py'
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    start_indices:
        Integers specifying the multistart to be plotted or int specifying
        up to which start index should be plotted
    references:
        Reference points that will be plotted along with the results

    Returns
    -------
    fvals:
        List of arrays of function values for each result
    offset_y:
        offset for the y-axis
    """
    min_val = np.inf
    fvals_all = []
    for result in results:
        fvals = np.array(result.optimize_result.fval)
        # todo: order of results plays a role
        start_indices = process_start_indices(result, start_indices)
        fvals = fvals[start_indices]
        # if none of the fvals are finite, set default value to zero as
        # np.nanmin will error for an empty array
        if np.isfinite(fvals).any():
            min_val = min(min_val, np.nanmin(fvals[np.isfinite(fvals)]))

        fvals_all.append(fvals)

    # if there are references, also account for those
    if references:
        min_val = min(min_val, np.nanmin([r['fval'] for r in references]))

    offset_y = process_offset_y(offset_y, scale_y, float(min_val))

    # return offsetted values
    return [fvals + offset_y for fvals in fvals_all], offset_y


def get_ordering_by_start_id(results: Sequence[Result]) -> List[int]:
    """Get an ordering of start IDs.

    The ordering is generated by taking the best function value for each
    start across all results, then sorting these best function values.
    This means that the minimum value of the waterfall plot is
    monotonically increasing.

    Parameters
    ----------
    results:
        The results.

    Returns
    -------
    The ordering.
    """
    if len(results) < 2:
        raise ValueError("Multiple result objects are required.")

    optimize_results = [r.optimize_result.list for r in results]

    # Check whether the same start IDs exist across all results.
    # Note that start IDs, and not the vectors themselves, are checked. This is
    # because, for example, when comparing hierarchical optimization to
    # standard optimization, the starts may be comparable even if the vectors
    # are not identical.
    ids0 = sorted([s.id for s in optimize_results[0]])
    for optimize_result in optimize_results[1:]:
        ids = sorted([s.id for s in optimize_result])
        if ids != ids0:
            raise ValueError("The start IDs of the results do not match.")

    start_fval_pairs = [
        (start.id, start.fval)
        for result in results
        for start in result.optimize_result.list
    ]

    sorted_start_fval_pairs = sorted(
        start_fval_pairs,
        key=lambda pair: pair[1],
    )

    ordering = []
    for start_id, _ in sorted_start_fval_pairs:
        if start_id in ordering:
            continue
        ordering.append(start_id)

    return ordering


def handle_options(ax, max_len_fvals, ref, y_limits, offset_y):
    """
    Apply post-plotting transformations to the axis object.

    Get the limits for the y-axis, plots the reference points, will do
    more at a later time point.

    Parameters
    ----------
    ax: matplotlib.Axes, optional
        Axes object to use.
    max_len_fvals: int
        maximum number of points
    ref: list, optional
        List of reference points for optimization results, containing at
        least a function value fval
    y_limits: float or ndarray, optional
        maximum value to be plotted on the y-axis, or y-limits
    offset_y:
        offset for the y-axis, if it is supposed to be in log10-scale

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # handle reference points
    for i_ref in ref:
        # plot reference point as line
        ax.plot(
            [0, max_len_fvals - 1],
            [i_ref.fval + offset_y, i_ref.fval + offset_y],
            '--',
            color=i_ref.color,
            label=i_ref.legend,
        )

        # create legend for reference points
        if i_ref.legend is not None:
            ax.legend()

    # handle y-limits
    ax = process_y_limits(ax, y_limits)

    return ax



from typing import Dict, Union


class ProfileOptions(dict):
    """
    Options for optimization based profiling.

    Parameters
    ----------
    default_step_size:
        Default step size of the profiling routine along the profile path
        (adaptive step lengths algorithms will only use this as a first guess
        and then refine the update).
    min_step_size:
        Lower bound for the step size in adaptive methods.
    max_step_size:
        Upper bound for the step size in adaptive methods.
    step_size_factor:
        Adaptive methods recompute the likelihood at the predicted point and
        try to find a good step length by a sort of line search algorithm.
        This factor controls step handling in this line search.
    delta_ratio_max:
        Maximum allowed drop of the posterior ratio between two profile steps.
    ratio_min:
        Lower bound for likelihood ratio of the profile, based on inverse
        chi2-distribution.
        The default 0.145 is slightly lower than the 95% quantile 0.1465 of a
        chi2 distribution with one degree of freedom.
    reg_points:
        Number of profile points used for regression in regression based
        adaptive profile points proposal.
    reg_order:
        Maximum degree of regression polynomial used in regression based
        adaptive profile points proposal.
    magic_factor_obj_value:
        There is this magic factor in the old profiling code which slows down
        profiling at small ratios (must be >= 0 and < 1).
    """

    def __init__(
        self,
        default_step_size: float = 0.01,
        min_step_size: float = 0.001,
        max_step_size: float = 1.0,
        step_size_factor: float = 1.25,
        delta_ratio_max: float = 0.1,
        ratio_min: float = 0.145,
        reg_points: int = 10,
        reg_order: int = 4,
        magic_factor_obj_value: float = 0.5,
    ):
        super().__init__()

        self.default_step_size = default_step_size
        self.min_step_size = min_step_size
        self.max_step_size = max_step_size
        self.ratio_min = ratio_min
        self.step_size_factor = step_size_factor
        self.delta_ratio_max = delta_ratio_max
        self.reg_points = reg_points
        self.reg_order = reg_order
        self.magic_factor_obj_value = magic_factor_obj_value

    def __getattr__(self, key):
        """Allow usage of keys like attributes."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    @staticmethod
    def create_instance(
        maybe_options: Union['ProfileOptions', Dict]
    ) -> 'ProfileOptions':
        """
        Return a valid options object.

        Parameters
        ----------
        maybe_options: ProfileOptions or dict
        """
        if isinstance(maybe_options, ProfileOptions):
            return maybe_options
        options = ProfileOptions(**maybe_options)
        return options



"""
Jax models interface.

Adds an interface for the construction of loss functions
incorporating jax models. This permits computation of derivatives using a
combination of objective based methods and jax based autodiff.
"""

import copy
from functools import partial
from typing import Callable, Sequence, Tuple

import numpy as np

from ...C import FVAL, GRAD, HESS, MODE_FUN, RDATAS, ModeType
from ..base import ObjectiveBase, ResultDict

try:
    import jax
    import jax.experimental.host_callback as hcb
    import jax.numpy as jnp
    from jax import custom_jvp, grad
except ImportError:
    raise ImportError(
        "Using a jax objective requires an installation of "
        "the python package jax. Please install jax via "
        "`pip install jax jaxlib`."
    )

# jax compatible (jittable) objective function using host callback, see
# https://jax.readthedocs.io/en/latest/jax.experimental.host_callback.html


@partial(custom_jvp, nondiff_argnums=(0,))
def _device_fun(obj: 'JaxObjective', x: jnp.array):
    """Jax compatible objective function execution using host callback.

    This function does not actually call the underlying objective function,
    but instead extracts cached return values. Thus it must only be called
    from within obj.call_unprocessed, and obj.cached_base_ret must be populated.

    Parameters
    ----------
    obj:
        The wrapped jax objective.
    x:
        jax computed input array.

    Note
    ----
    This function should rather be implemented as class method of JaxObjective,
    but this is not possible at the time of writing as this is not supported
    by signature inspection in the underlying bind call.
    """
    return hcb.call(
        obj.cached_fval,
        x,
        result_shape=jax.ShapeDtypeStruct((), np.float64),
    )


@partial(custom_jvp, nondiff_argnums=(0,))
def _device_fun_grad(obj: 'JaxObjective', x: jnp.array):
    """Jax compatible objective gradient execution using host callback.

    This function does not actually call the underlying objective function,
    but instead extracts cached return values. Thus it must only be called
    from within obj.call_unprocessed and obj.cached_base_ret must be populated.

    Parameters
    ----------
    obj:
        The wrapped jax objective.
    x:
        jax computed input array.

    Note
    ----
    This function should rather be implemented as class method of JaxObjective,
    but this is not possible at the time of writing as this is not supported
    by signature inspection in the underlying bind call.
    """
    return hcb.call(
        obj.cached_grad,
        x,
        result_shape=jax.ShapeDtypeStruct(
            obj.cached_base_ret[GRAD].shape,  # bootstrap from cached value
            np.float64,
        ),
    )


def _device_fun_hess(obj: 'JaxObjective', x: jnp.array):
    """Jax compatible objective Hessian execution using host callback.

    This function does not actually call the underlying objective function,
    but instead extracts cached return values. Thus it must only be called
    from within obj.call_unprocessed and obj.cached_base_ret must be populated.

    Parameters
    ----------
    obj:
        The wrapped jax objective.
    x:
        jax computed input array.

    Note
    ----
    This function should rather be implemented as class method of JaxObjective,
    but this is not possible at the time of writing as this is not supported
    by signature inspection in the underlying bind call.
    """
    return hcb.call(
        obj.cached_hess,
        x,
        result_shape=jax.ShapeDtypeStruct(
            obj.cached_base_ret[HESS].shape,  # bootstrap from cached value
            np.float64,
        ),
    )


# define custom jvp for device_fun & device_fun_grad to enable autodiff, see
# https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html


@_device_fun.defjvp
def _device_fun_jvp(
    obj: 'JaxObjective', primals: jnp.array, tangents: jnp.array
):
    """JVP implementation for device_fun."""
    (x,) = primals
    (x_dot,) = tangents
    return _device_fun(obj, x), _device_fun_grad(obj, x).dot(x_dot)


@_device_fun_grad.defjvp
def _device_fun_grad_jvp(
    obj: 'JaxObjective', primals: jnp.array, tangents: jnp.array
):
    """JVP implementation for device_fun_grad."""
    (x,) = primals
    (x_dot,) = tangents
    return _device_fun_grad(obj, x), _device_fun_hess(obj, x).dot(x_dot)


class JaxObjective(ObjectiveBase):
    """Objective function that combines pypesto objectives with jax functions.

    The generated objective function will evaluate objective(jax_fun(x)).

    Parameters
    ----------
    objective:
        pyPESTO objective
    jax_fun:
        jax function (not jitted) that computes input to the pyPESTO objective
    """

    def __init__(
        self,
        objective: ObjectiveBase,
        jax_fun: Callable,
        x_names: Sequence[str] = None,
    ):
        if not isinstance(objective, ObjectiveBase):
            raise TypeError('objective must be an ObjectiveBase instance')
        if not objective.check_mode(MODE_FUN):
            raise NotImplementedError(
                f'objective must support mode={MODE_FUN}'
            )
        super().__init__(x_names)
        self.base_objective = objective

        self.jax_fun = jax_fun

        # would be cleaner to also have this as class method, but not supported
        # by signature inspection in bind call.
        def jax_objective(x):
            # device fun doesn't actually need the value of y, but we need to
            # compute this here for autodiff to work
            y = jax_fun(x)
            return _device_fun(self, y)

        # jit objective & derivatives (not integrated)
        self.jax_objective = jax.jit(jax_objective)
        self.jax_objective_grad = jax.jit(grad(jax_objective))
        self.jax_objective_hess = jax.jit(jax.hessian(jax_objective))

        # jit input function
        self.infun = jax.jit(self.jax_fun)

        # temporary storage for evaluation results of objective
        self.cached_base_ret: ResultDict = {}

    def cached_fval(self, _):
        """Return cached function value."""
        return self.cached_base_ret[FVAL]

    def cached_grad(self, _):
        """Return cached gradient."""
        return self.cached_base_ret[GRAD]

    def cached_hess(self, _):
        """Return cached Hessian."""
        return self.cached_base_ret[HESS]

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return mode == MODE_FUN

    def check_sensi_orders(self, sensi_orders, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        if not self.check_mode(mode):
            return False
        else:
            return self.base_objective.check_sensi_orders(sensi_orders, mode)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        **kwargs,
    ) -> ResultDict:
        """
        See `ObjectiveBase` for more documentation.

        Main method to overwrite from the base class. It handles and
        delegates the actual objective evaluation.
        """
        # derivative computation in jax always requires lower order
        # derivatives, see jvp rules for device_fun and device_fun_grad.
        if 2 in sensi_orders:
            sensi_orders = (0, 1, 2)
        elif 1 in sensi_orders:
            sensi_orders = (0, 1)
        else:
            sensi_orders = (0,)

        # this computes all the results from the inner objective, rendering
        # them accessible as cached values for device_fun, etc.
        set_return_dict, return_dict = (
            'return_dict' in kwargs,
            kwargs.pop('return_dict', False),
        )
        self.cached_base_ret = self.base_objective(
            self.infun(x), sensi_orders, mode, return_dict=True, **kwargs
        )
        if set_return_dict:
            kwargs['return_dict'] = return_dict
        ret = {}
        if RDATAS in self.cached_base_ret:
            ret[RDATAS] = self.cached_base_ret[RDATAS]
        if 0 in sensi_orders:
            ret[FVAL] = float(self.jax_objective(x))
        if 1 in sensi_orders:
            ret[GRAD] = self.jax_objective_grad(x)
        if 2 in sensi_orders:
            ret[HESS] = self.jax_objective_hess(x)

        return ret

    def __deepcopy__(self, memodict=None):
        other = JaxObjective(
            copy.deepcopy(self.base_objective),
            copy.deepcopy(self.jax_fun),
            copy.deepcopy(self.x_names),
        )

        return other



"""
Storage
=======

Saving and loading traces and results objects.
"""

from .auto import autosave
from .hdf5 import write_array
from .read_from_hdf5 import (
    OptimizationResultHDF5Reader,
    ProblemHDF5Reader,
    ProfileResultHDF5Reader,
    SamplingResultHDF5Reader,
    load_objective_config,
    read_result,
)
from .save_to_hdf5 import (
    OptimizationResultHDF5Writer,
    ProblemHDF5Writer,
    ProfileResultHDF5Writer,
    SamplingResultHDF5Writer,
    write_result,
)



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Callable, List, Sequence, Tuple, Union

import numpy as np

from ..C import (
    AMICI_LLH,
    AMICI_SIGMAY,
    AMICI_STATUS,
    AMICI_SX,
    AMICI_SY,
    AMICI_T,
    AMICI_X,
    AMICI_Y,
    CSV,
    H5,
    MODE_FUN,
    OUTPUT,
    OUTPUT_IDS,
    OUTPUT_SENSI,
    OUTPUT_SIGMAY,
    OUTPUT_WEIGHT,
    PARAMETER_IDS,
    RDATAS,
    TIMEPOINTS,
    ModeType,
)

if TYPE_CHECKING:
    from ..objective import AmiciObjective

from ..result import PredictionResult


class AmiciPredictor:
    """
    Do forward simulations/predictions for an AMICI model.

    The user may supply post-processing methods.
    If post-processing methods are supplied, and a gradient of the prediction
    is requested, then the sensitivities of the AMICI model must also be
    post-processed. There are no checks here to ensure that the sensitivities
    are correctly post-processed, this is explicitly left to the user.
    There are also no safeguards if the postprocessor routines fail. This may
    happen if, e.g., a call to Amici fails, and no timepoints, states or
    observables are returned. As the AmiciPredictor is agnostic about the
    dimension of the postprocessor and also the dimension of the postprocessed
    output, these checks are also left to the user. An example for such a check
    is provided in the default output (see _default_output()).
    """

    def __init__(
        self,
        amici_objective: AmiciObjective,
        amici_output_fields: Union[Sequence[str], None] = None,
        post_processor: Union[Callable, None] = None,
        post_processor_sensi: Union[Callable, None] = None,
        post_processor_time: Union[Callable, None] = None,
        max_chunk_size: Union[int, None] = None,
        output_ids: Union[Sequence[str], None] = None,
        condition_ids: Union[Sequence[str], None] = None,
    ):
        """
        Initialize predictor.

        Parameters
        ----------
        amici_objective:
            An objective object, which will be used to get model simulations
        amici_output_fields:
            keys that exist in the return data object from AMICI, which should
            be available for the post-processors
        post_processor:
            A callable function which applies postprocessing to the simulation
            results and possibly defines different outputs than those of
            the amici model. Default are the observables
            (`pypesto.C.AMICI_Y`) of the AMICI model. This
            method takes a list of dicts (with the returned
            fields `pypesto.C.AMICI_T`,
            `pypesto.C.AMICI_X`, and
            `pypesto.C.AMICI_Y` of the AMICI ReturnData
            objects) as input. Safeguards for, e.g., failure of AMICI are left
            to the user.
        post_processor_sensi:
            A callable function which applies postprocessing to the
            sensitivities of the simulation results. Defaults to the
            observable sensitivities of the AMICI model.
            This method takes a list of dicts (with the returned fields
            `pypesto.C.AMICI_T`,
            `pypesto.C.AMICI_X`,
            `pypesto.C.AMICI_Y`,
            `pypesto.C.AMICI_SX`, and
            `pypesto.C.AMICI_SY` of the AMICI ReturnData
            objects) as input. Safeguards for, e.g., failure of AMICI are left
            to the user.
        post_processor_time:
            A callable function which applies postprocessing to the timepoints
            of the simulations. Defaults to the timepoints of the amici model.
            This method takes a list of dicts (with the returned field
            `pypesto.C.AMICI_T` of the amici ReturnData
            objects) as input. Safeguards for, e.g., failure of AMICI are left
            to the user.
        max_chunk_size:
            In some cases, we don't want to compute all predictions at once
            when calling the prediction function, as this might not fit into
            the memory for large datasets and models.
            Here, the user can specify a maximum chunk size of conditions,
            which should be simulated at a time.
            Defaults to None, meaning that all conditions will be simulated.
        output_ids:
            IDs of outputs, as post-processing allows the creation of
            customizable outputs, which may not coincide with those from
            the AMICI model (defaults to AMICI observables).
        condition_ids:
            List of identifiers for the conditions of the edata objects of the
            amici objective, will be passed to the PredictionResult at call.
        """
        # save settings and objective
        self.amici_objective = amici_objective
        self.max_chunk_size = max_chunk_size
        self.post_processor = post_processor
        self.post_processor_sensi = post_processor_sensi
        self.post_processor_time = post_processor_time
        self.condition_ids = condition_ids

        # If the user takes care of everything we can skip default readouts
        self.skip_default_outputs = False
        if (
            post_processor is not None
            and post_processor_sensi is not None
            and post_processor_time is not None
        ):
            self.skip_default_outputs = True

        self.output_ids = output_ids
        if output_ids is None:
            self.output_ids = amici_objective.amici_model.getObservableIds()

        if amici_output_fields is None:
            amici_output_fields = [
                AMICI_STATUS,
                AMICI_T,
                AMICI_Y,
                AMICI_SY,
                AMICI_X,
                AMICI_SX,
            ]
        self.amici_output_fields = amici_output_fields

    def __call__(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...] = (0,),
        mode: ModeType = MODE_FUN,
        output_file: str = '',
        output_format: str = CSV,
        include_llh_weights: bool = False,
        include_sigmay: bool = False,
    ) -> PredictionResult:
        """
        Call the predictor.

        Simulate a model for a certain prediction function.
        This method relies on the AmiciObjective, which is underlying, but
        allows the user to apply any post-processing of the results, the
        sensitivities, and the timepoints.

        Parameters
        ----------
        x:
            The parameters for which to evaluate the prediction function.
        sensi_orders:
            Specifies which sensitivities to compute, e.g. (0,1) -> fval, grad.
        mode:
            Whether to compute function values or residuals.
        output_file:
            Path to an output file.
        output_format:
            Either 'csv', 'h5'. If an output file is specified, this routine
            will return a csv file, created from a DataFrame, or an h5 file,
            created from a dict.
        include_llh_weights:
            Boolean whether weights should be included in the prediction.
            Necessary for weighted means of Ensembles.
        include_sigmay:
            Boolean whether standard deviations should be included in the
            prediction output. Necessary for evaluation of weighted means
            of Ensembles.

        Returns
        -------
        results:
            PredictionResult object containing timepoints, outputs, and
            output_sensitivities if requested
        """
        # sanity check for output
        if 2 in sensi_orders:
            raise Exception(
                'Prediction simulation does currently not support '
                'second order output.'
            )
        # add llh and sigmay to amici output fields if requested
        if include_llh_weights and AMICI_LLH not in self.amici_output_fields:
            self.amici_output_fields.append(AMICI_LLH)
        if include_sigmay and AMICI_SIGMAY not in self.amici_output_fields:
            self.amici_output_fields.append(AMICI_SIGMAY)

        # simulate the model and get the output
        (
            timepoints,
            outputs,
            outputs_sensi,
            outputs_weight,
            outputs_sigmay,
        ) = self._get_outputs(
            x, sensi_orders, mode, include_llh_weights, include_sigmay
        )

        # group results by condition, prepare PredictionConditionResult output
        condition_results = []
        # timepoints, outputs, outputs_sensi, outputs_sigmay and
        # outputs_weight are lists with the number of simulation conditions.
        # While everything else is optional, timepoints must exist,
        # so we use this as a dummy
        n_cond = len(timepoints)
        for i_cond in range(n_cond):
            result = {
                TIMEPOINTS: timepoints[i_cond],
                OUTPUT_IDS: self.output_ids,
                PARAMETER_IDS: self.amici_objective.x_names,
            }
            if outputs:
                result[OUTPUT] = outputs[i_cond]
            if outputs_sensi:
                result[OUTPUT_SENSI] = outputs_sensi[i_cond]
            if outputs_weight:
                result[OUTPUT_WEIGHT] = outputs_weight[i_cond]
            if outputs_sigmay:
                result[OUTPUT_SIGMAY] = outputs_sigmay[i_cond]

            condition_results.append(result)
        # create result object
        results = PredictionResult(
            condition_results, condition_ids=self.condition_ids
        )

        # Should the results be saved to a file?
        if output_file:
            # Do we want a pandas dataframe like format?
            if output_format == CSV:
                results.write_to_csv(output_file=output_file)
            # Do we want an h5 file?
            elif output_format == H5:
                results.write_to_h5(output_file=output_file)
            else:
                raise Exception(
                    f'Call to unknown format {output_format} for '
                    f'output of pyPESTO prediction.'
                )

        # return dependent on sensitivity order
        return results

    def _get_outputs(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType = MODE_FUN,
        include_llh_weights: bool = False,
        include_sigmay: bool = False,
    ) -> Tuple[List, List, List]:
        """
        Split the calls to amici into smaller chunks.

        Too large ReturnData objects from amici including many simulations
        can be problematic in terms of memory.

        Parameters
        ----------
        x:
            The parameters for which to evaluate the prediction function.
        sensi_orders:
            Specifies which sensitivities to compute, e.g. (0,1) -> fval, grad.
        mode:
            Whether to compute function values or residuals.

        Returns
        -------
        timepoints:
            List of np.ndarrays, every entry includes the output timepoints of
            the respective condition
        outputs:
            List of np.ndarrays, every entry includes the postprocessed outputs
            of the respective condition
        outputs_sensi:
            List of np.ndarrays, every entry includes the postprocessed output
            sensitivities of the respective condition
        include_llh_weights:
            Boolean whether weights should be included in the prediction.
            Necessary for weighted means of Ensembles.
        include_sigmay:
            Boolean whether standard deviations should be included in the
            prediction output. Necessary for evaluation of weighted means
            of Ensembles.
        """
        # Do we have a maximum number of simulations allowed?
        n_edatas = len(self.amici_objective.edatas)
        if self.max_chunk_size is None:
            # simulate all conditions at once
            n_simulations = 1
        else:
            # simulate only a subset of conditions
            n_simulations = int(np.ceil(n_edatas / self.max_chunk_size))

        # prepare result
        amici_outputs = []

        for i_sim in range(n_simulations):
            # slice out the conditions we actually want
            if self.max_chunk_size is None:
                ids = slice(0, n_edatas)
            else:
                ids = slice(
                    i_sim * self.max_chunk_size,
                    min((i_sim + 1) * self.max_chunk_size, n_edatas),
                )

            # call amici
            self._wrap_call_to_amici(
                amici_outputs=amici_outputs,
                x=x,
                sensi_orders=sensi_orders,
                parameter_mapping=self.amici_objective.parameter_mapping[ids],
                edatas=self.amici_objective.edatas[ids],
                mode=mode,
            )

        def _default_output(amici_outputs):
            """
            Create default output of prediction.

            Equals to observables of AMICI model. We need to check that call
            to AMICI was successful (status == 0), before writing the output.
            """
            amici_nt = [
                len(edata.getTimepoints())
                for edata in self.amici_objective.edatas
            ]
            amici_ny = len(self.output_ids)
            amici_np = len(self.amici_objective.x_names)

            outputs = []
            outputs_sensi = []
            outputs_weights = []
            outputs_sigmay = []
            timepoints = [
                amici_output[AMICI_T]
                if amici_output[AMICI_STATUS] == 0
                else np.full((amici_nt[i_condition],), np.nan)
                for i_condition, amici_output in enumerate(amici_outputs)
            ]
            # add outputs and sensitivities if requested
            if 0 in sensi_orders:
                outputs = [
                    amici_output[AMICI_Y]
                    if amici_output[AMICI_STATUS] == 0
                    else np.full((amici_nt[i_condition], amici_ny), np.nan)
                    for i_condition, amici_output in enumerate(amici_outputs)
                ]
            if 1 in sensi_orders:
                outputs_sensi = [
                    amici_output[AMICI_SY]
                    if amici_output[AMICI_STATUS] == 0
                    else np.full(
                        (amici_nt[i_condition], amici_np, amici_ny), np.nan
                    )
                    for i_condition, amici_output in enumerate(amici_outputs)
                ]
            # add likelihood as weights if requested
            if include_llh_weights:
                outputs_weights = [
                    amici_output[AMICI_LLH]
                    if amici_output[AMICI_STATUS] == 0
                    else np.nan
                    for i_condition, amici_output in enumerate(amici_outputs)
                ]
            # add standard deviations if requested
            if include_sigmay:
                outputs_sigmay = [
                    amici_output[AMICI_SIGMAY]
                    if amici_output[AMICI_STATUS] == 0
                    else np.full((1, amici_ny), np.nan)
                    for i_condition, amici_output in enumerate(amici_outputs)
                ]

            return (
                timepoints,
                outputs,
                outputs_sensi,
                outputs_weights,
                outputs_sigmay,
            )

        outputs_weights = []
        outputs_sigmay = []
        # Get default output
        if not self.skip_default_outputs:
            (
                timepoints,
                outputs,
                outputs_sensi,
                outputs_weights,
                outputs_sigmay,
            ) = _default_output(amici_outputs)

        # postprocess (use original Amici outputs)
        if self.post_processor is not None:
            outputs = self.post_processor(amici_outputs)
        if self.post_processor_sensi is not None:
            outputs_sensi = self.post_processor_sensi(amici_outputs)
        if self.post_processor_time is not None:
            timepoints = self.post_processor_time(amici_outputs)

        return (
            timepoints,
            outputs,
            outputs_sensi,
            outputs_weights,
            outputs_sigmay,
        )

    def _wrap_call_to_amici(
        self, amici_outputs, x, sensi_orders, mode, parameter_mapping, edatas
    ):
        """
        Encapsulate the call to amici.

        This allows to use variable scoping as a mean to clean up the memory
        after calling amici, which is beneficial if large models with large
        datasets are used.
        """
        chunk = self.amici_objective(
            x=x,
            sensi_orders=sensi_orders,
            mode=mode,
            parameter_mapping=parameter_mapping,
            edatas=edatas,
            return_dict=True,
        )
        for rdata in chunk[RDATAS]:
            amici_outputs.append(
                {
                    output_field: deepcopy(rdata[output_field])
                    for output_field in self.amici_output_fields
                }
            )
        del chunk
